/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon
 */

#ifndef _CONNECTION_H_
#define _CONNECTION_H_

#include "ResultAndWarnings.hpp"
#include "ConnectionParameters.hpp"
#include "DriverResultSet.hpp"

#include "CriticalSection.hpp"
#include "CarobException.hpp"

#include <list>

namespace {
const int32_t ProtocolVersion = (static_cast<int32_t>(8) /* major */ << 16)
                                                    + 10 /* minor */;
}

namespace CarobNS {

/** "Controller Ready" magic number. */
#define ControllerPrompt                        0x1FA1501F
/** Command prefix sent before each command */
#define CommandPrefix                           0xB015CA

#define Ping                                    -1
#define StatementExecuteQuery                   0
#define StatementExecuteUpdate                  1
#define StatementExecuteUpdateWithKeys          2
//#define CallableStatementExecuteQuery           3
//#define CallableStatementExecuteUpdate          4
//#define CallableStatementExecute                5
#define RetrieveExecuteQueryResult              10
#define RetrieveExecuteUpdateResult             11
#define RetrieveExecuteResult                   12
#define RetrieveExecuteUpdateWithKeysResult     13
#define RetrieveCommitResult                    14
#define RetrieveRollbackResult                  15
#define StatementExecute                        6
#define Begin                                   20
#define Commit                                  21
#define Rollback                                22
//#define SetNamedSavepoint                       23
//#define SetUnnamedSavepoint                     24
//#define ReleaseSavepoint                        25
//#define RollbackToSavepoint                     26
#define Close                                   30
// We don't pool connections => Reset is not used
//#define Reset                                   31
#define FetchNextResultSetRows                  32
#define CloseRemoteResultSet                    33
#define RestoreConnectionState                  34
// deprecated, see CAROB-95
// SetAutoCommit                           35
//#define ConnectionGetCatalog                    36
//#define ConnectionGetCatalogs                   37
//#define ConnectionSetCatalog                    38
//#define SetTransactionIsolation                 39
#define ConnectionGetWarnings                   40
#define ConnectionClearWarnings                 41
#define SetReadOnly                             42
//#define GetVirtualDatabaseName                  50
//#define GetControllerVersionNumber              51
//#define DatabaseMetaDataGetTables               52
//#define DatabaseMetaDataGetColumns              53
//#define DatabaseMetaDataGetPrimaryKeys          54
//#define DatabaseMetaDataGetProcedures           55
//#define DatabaseMetaDataGetProcedureColumns     56
//#define DatabaseMetaDataGetTableTypes           58
//#define DatabaseMetaDataGetTablePrivileges      59
//#define DatabaseMetaDataGetSchemas              60
//#define DatabaseMetaDataGetDatabaseProductName  61
//#define DatabaseMetaDataGetAttributes           62
//#define DatabaseMetaDataGetBestRowIdentifier    63
//#define DatabaseMetaDataGetColumnPrivileges     64
//#define DatabaseMetaDataGetCrossReference       65
//#define DatabaseMetaDataGetExportedKeys         66
//#define DatabaseMetaDataGetImportedKeys         67
//#define DatabaseMetaDataGetIndexInfo            68
//#define DatabaseMetaDataGetSuperTables          69
//#define DatabaseMetaDataGetSuperTypes           70
//#define DatabaseMetaDataGetTypeInfo             71
//#define DatabaseMetaDataGetUDTs                 72
//#define DatabaseMetaDataGetVersionColumns       73
//#define DatabaseStaticMetadata                  80
#define PreparedStatementGetMetaData            81

class ControllerInfo;
class DriverSocket;
class ConnectionParameters;
class AbstractControllerPool;
class Statement;
class ParameterStatement;
class Request;
class RequestWithResultSetParameters;
class DriverResultSet;

/**
 * Exception on authentication. Send on connection (user/pass)
 */
class AuthenticationException : public DriverException
{
public:
  /**
   * Constructs a AuthenticationException with the given message
   * @param s error message of the exception
   */
  AuthenticationException(std::wstring s) : DriverException(s, L"28000") {}
};


/**
 * This class implements the communication protocol to the Controller.
 */
class Connection
{
// These functions need access to getDriverSocket()
friend DriverResultSet::DriverResultSet(Connection* conn)
    throw (ProtocolException, SocketIOException, NotImplementedException,
    UnexpectedException);
friend bool DriverResultSet::next() throw (SocketIOException,
    ControllerException, ProtocolException, BackendException,
    UnexpectedException);

public:
  /**
   * Creates a new connection and connects to a controller with the given
   * parameters
   * @param prms controller, base, user, etc. to use for connecting
   * @throw DriverException if the ConnectPolicy given in parameters is unknown
   * @throw ConnectionException if the requested controller could not be
   *        reached
   * @throw AuthenticationException if a login error occured, or if an 
   *        IO exception occured during authentication (premature close of
   *        connection)
   */
  Connection(ConnectionParameters& prms) throw (DriverException,
      ConnectionException, AuthenticationException, UnexpectedException);

  /**
   * Closes the connection and destroys its socket
   */
  virtual ~Connection();

	/**
	 * Before sending a command code, checks that the controller is
	 * synchronized/ready to accept it. Then sends it.
	 * 
	 * @param socket on which to read and write
	 * @param command to send
	 * @throws IOException on socket error
	 * @throws ProtocolException on protocol corruption
	 */
	static void       sendCommand(const DriverSocket& socket, int command)
                        throw (SocketIOException, ProtocolException,
                        UnexpectedException);
  /**
   * If a connection is in auto-commit mode, then all its SQL statements will be
   * executed and committed as individual transactions. Otherwise, its SQL
   * statements are grouped into transactions that are terminated by either
   * {@link #commit()}or {@link #rollback()}. By default, new connections are
   * in auto-commit mode. The commit occurs when the statement completes or the
   * next execute occurs, whichever comes first. In the case of statements
   * returning a <code>ResultSet</code>, the statement completes when the
   * last row of the <code>ResultSet</code> has been retrieved or the
   * <code>ResultSet</code> has been closed.
   * @param autoCommit <code>true</code> enables auto-commit;
   *          <code>false</code> disables it
   * @throws SocketIOException if a database access error occurs
   */
  void              setAutoCommit(bool autoCommit)
                        throw (SocketIOException, BackendException, 
                        ControllerException, ProtocolException,
                        ConnectionException, UnexpectedException);
  /**
   * Makes all changes made since the previous commit/rollback permanent and
   * releases any database locks currently held by the <code>Connection</code>.
   * This method should only be used when auto-commit has been disabled. (If
   * <code>autoCommit</code>== <code>true</code>, then we throw a
   * DriverException).
   * 
   * @throws DriverException  if connection is in autocommit mode
   * @see Connection#setAutoCommit(boolean)
   */
  void              commit() throw (SocketIOException, DriverException,
                        ProtocolException, ConnectionException,
                        UnexpectedException);
  /**
   * Drops all changes made since the previous commit/rollback and releases any
   * database locks currently held by this connection. If the connection was in
   * autocommit mode, we throw a DriverException.
   * 
   * @exception DriverException if the connection is in autocommit mode
   * @see Connection#commit()
   */
  void              rollback() throw (SocketIOException, DriverException,
                        ProtocolException, ConnectionException,
                        UnexpectedException);
  /**
   * SQL statements without parameters are normally executed using
   * <code>Statement</code> objects. If the same SQL statement is executed
   * many times, it is more efficient to use a <code>ParameterStatement</code>.
   * The <code>ResultSet</code> will be
   * <code>TYPE_FORWARD_ONLY</code>/<code>CONCUR_READ_ONLY</code>.
   *
   * @return a new <code>Statement</code> object
   */
  Statement*        createStatement() throw (SocketIOException,
                        UnexpectedException);
  /**
   * Creates a parameter statement.
   * @param query a sql skeleton query
   * @return a new <code>ParameterStatement</code> object
   */
  ParameterStatement* createParameterStatement(const std::wstring &query) 
                        throw (SocketIOException, UnexpectedException);
  /**
   * Deletes given statement.
   * Only way to delete a statement (private dtor) so connection will keep the
   * control of its statement and can remove it from its statement list.
   * If the given statement pointer is NULL, don't do anything
   * @param stPtr pointer to the statement to destroy
   * @throw DriverException if the given statement is not in our list
   */
  void              deleteStatement(Statement* stPtr) throw (DriverException,
                        UnexpectedException);
  /**
   * Performs a read request and returns the reply.
   * 
   * @param request the read request with result set parameters to execute
   * @return a pointer to the resulting <code>DriverResultSet</code>
   * @throw SocketIOException if an error occurs on the socket
   */
  DriverResultSet*  statementExecuteQuery(RequestWithResultSetParameters& request)
                        throw (SocketIOException, BackendException, 
                        ControllerException, ProtocolException,
                        NotImplementedException, ConnectionException,
                        UnexpectedException);
  /**
   * Performs a write request and return the number of rows affected.
   * 
   * @param r the write request to execute
   * @return number of rows affected, -1 if an error occured
   * @throws SocketIOException if an error occurs
   */
  ResultAndWarnings statementExecuteUpdate(Request& r)
                        throw (SocketIOException, BackendException, 
                        ControllerException, ProtocolException,
                        ConnectionException, UnexpectedException);
  /**
   * Calls a request that returns a list of results and the optionnal warnings.
   * This list is returned as a copy (ie. not a pointer) because it contains
   * only lightweight elements (pointers or ints), so there is no heavy
   * ResultSet copy.
   * 
   * @param request the request to execute
   * @return a <code>list</code> of <code>ResultSetOrUpdateCount</code> elements
   * @throws SocketIOException if an error occurs
   */
  ResultAndWarnings statementExecute(RequestWithResultSetParameters& request)
                                        throw (SocketIOException,
                                        BackendException, ControllerException,
                                        ProtocolException, ConnectionException,
                                        UnexpectedException);
  /**
   * Performs a write request and returns the auto-generated keys
   * 
   * @param request the write request to execute
   * @return a <code>list</code> of 2 <code>ResultSetOrUpdateCount</code>
   *          elements in which the first element is the update count returned
   *          by the executeUpdate execution, the second element being a
   *          DriverResultSet containing the autogenerated keys
   * @throws SocketIOException if an error occurs
   */
  ResultAndWarnings statementExecuteUpdateWithKeys(
                                        Request &request)
                                        throw (SocketIOException,
                                        BackendException,ControllerException,
                                        ProtocolException, ConnectionException,
                                        UnexpectedException);
  /**
   * Closes the remote ResultSet given its cursor name.
   * 
   * @param cursorName cursor name of the ResultSet to close
   * @throw SocketIOException if an error occurs on the socket
   */
  void              closeRemoteResultSet(std::wstring cursorName)
                        throw (SocketIOException, BackendException,
                        ControllerException, ProtocolException,
                        UnexpectedException);
  /**
   * Tests to see if the connection is in read only Mode. Note that we cannot
   * really put the database in read only mode, but we pretend we can by
   * returning the value of the <code>readOnly</code> flag.
   * 
   * @return <code>true</code> if the connection is read only
   */
  bool              isReadOnly() { return readOnly; }
  /**
   * A connection can be put in read-only mode as a hint to enable database
   * optimizations
   * 
   * @param readOnlyPrm  <code>true</code> enables read-only mode;
   *                  <code>false</code> disables it
   * @throw DriverException if transaction executes write requests
   */
  void              setReadOnly(bool readOnlyPrm) throw (SocketIOException,
                        DriverException, UnexpectedException);
  /**
   * Sends "FetchNextResultSetRows" command to controller. Throws a
   * ControllerException if controller returns an error; else returns void and
   * lets the caller receive its new rows by itself.
   * 
   * @param cursorName name of the ResultSet cursor
   * @param fetchSize number of rows to fetch
   * @throws ControllerException if an error occurs
   */
  void              tryFetchNext(const std::wstring& cursorName, int fetchSize)
                        throw (ControllerException, SocketIOException,
                        ProtocolException, BackendException,
                        UnexpectedException);
  /**
   * Returns the metadata of the sql template as a resultset (which will contain
   * the metadata).
   * @param sqlTemplate sql template of the PreparedStatement
   */
  DriverResultSet*  preparedStatementGetMetaData(const std::wstring& sqlTemplate)
                          throw (SocketIOException, BackendException,
                          ProtocolException, ControllerException,
                          NotImplementedException, ConnectionException,
                          UnexpectedException);
  /**
   * Returns the first warning reported by calls on this connection. Subsequent
   * warnings will be chained to this SQLWarning<br>
   * <B>Note: </B> If the 'persistent connections' option is set to false, this
   * function will always return null.
   * 
   * @return the first SQLWarning or null
   * @exception DriverException if a database access error occurs or this method is
   *              called on a closed connection
   */
  SQLWarning*       getWarnings() throw (DriverException, UnexpectedException);
  /**
   * After this call, <code>getWarnings()</code> returns <code>null</code>
   * until a new call to getWarnings() on this connection.
   */
  void              clearWarnings() throw (DriverException, UnexpectedException);
  /**
   * Utility function to convert the given chain of backendException to a chain
   * of SQLWarnings
   * 
   * @param bde exception chain to convert
   */
  static SQLWarning* convertToSQLWarnings(BackendException* bde);

    
protected:
  /**
   * Starts a connection by sending all connection/authentication informations.
   * Warning: this function only sends the connection infos, but there is no
   * guaranty that we are connected afterwards. {@link #finalizeConnect()} will
   * do the job of checking and finishing connection establishment.
   * @return true if the connection initialization succeeds
   * @throw DriverException if the connect policy is NULL
   * @throw ConnectionException if the requested controller could not be
   *        reached
   * @throw AuthenticationException if a login error occured, or if an 
   *        IO exception occured during authentication (premature close of
   *        connection)
   */
  void                initConnection()
                          throw (DriverException, ConnectionException,
                          AuthenticationException, UnexpectedException);
  /**
   * Reads the controller acknowledgements and misc parameters that finish 
   * connection. {@link #initConnection()} must be called before this
   * @throw ConnectionException if the virtual database is not available on the
   *        controller
   * @throw AuthenticationException if the controller did not acknowledge or in
   *        case of i/o exception during finalization (premature close of
   *        connection)s
   */
  void                finalizeConnect() throw (ConnectionException,
                          AuthenticationException, UnexpectedException);

  /**
   * Deletes all statements created by this connection and closes the socket
   * to the connected controller.
   * If an error occurs while closing, just logs the error and returns false,
   * but does not throw any exception.
   * @return true if the connection has been closed without errors
   */
  bool                close();
  /**
   * Closes the socket by sending appropriate commands to the controller.
   * If an error occurs while closing, just logs the error and returns false,
   * but does not throw any exception.
   * @return true if the connection has been closed without errors
   */
  bool                closeSocket();
  /**
   * Serializes a write request on the output stream by sending only the needed
   * parameters to reconstruct it on the controller
   * @param request the write request to send
   * @throw SocketIOException when an error occured on the stream
   */
  void                writeRequestOnStream(const Request& request)
                          throw (SocketIOException, UnexpectedException);
  
  /**
   * Handy method to call before any send/receive operation: throws and 
   * exception if the socket is not valid
   * @throw SocketIOException
   */
  void                checkIfConnected() const throw (SocketIOException,
                          UnexpectedException);

  /**
   * Gets the driver socket for i/o operations
   * @return reference to this connection socket
   */
  const DriverSocket& getDriverSocket() { return *driverSocketPtr; }
                        
private:
  /** This connection socket pointer */
  DriverSocket*       driverSocketPtr;
  /** true if the connection is closed (or not opened...) */
  bool                isClosed;

  /** Commit mode of the connection (<code>true</code>= automatic). */
  bool                autoCommit;
  /** Is the connection in read-only mode ? */
  bool                readOnly;
  /** Has a write request been executed in the current transaction? */
  bool                writeExecutedInTransaction;
  /** Transaction identifier */
  int64_t             transactionId;
  /** Function-specific mutexes */
  CriticalSection     connectionCS;
  /**
   * Flag to check if a new transaction must be started before executing any
   * statement
   */
  bool                mustBeginTransaction;
  /**
   * True if the connection must be persisted on all cluster backends even when
   * autoCommit=true
   */
  bool                persistent_connection;

  /** Persistent connection identifier if persistentConnection is true */
  int64_t             persistent_connection_id;
  /** True if the controller should retrieve SQL warnings */
  bool                retrieve_sql_warnings;
  /** Pointer to the first warning attached to this connection */
  SQLWarning*         first_warning;
  /** Controller pool providing controllers to connect to */
  AbstractControllerPool& controller_pool;
  /** The parameters of this connection */
  ConnectionParameters parameters;
  /** The controller we are connected to */
  ControllerInfo      connected_controller;
  /**
   * List of pointers to the [parameter]statements created by this connection.
   * Kept in order to destroy them when closing connection
   */
  std::list<Statement*> created_statements;
  /**
   * Begins a new transaction if needed (<code>mustBeginTransaction</code> is
   * set to <code>true</code>).
   */
  void                beginTransactionIfNeeded() throw (SocketIOException,
                          BackendException, ControllerException, 
                          ProtocolException, ConnectionException,
                          UnexpectedException);

  /**
   * Actually commits. Code shared with setAutoCommit(true).
   */
  void                doCommit() throw (SocketIOException, DriverException,
                          ProtocolException, ConnectionException,
                          UnexpectedException);

  /**
   * Set the autocommit mode and read-only status on this request.
   * @param request The request to set
   */
  void                setConnectionParametersOnRequest(Request &request);

  void                throwExceptionIfClosed(const std::wstring& msg)
                          throw (DriverException, UnexpectedException);
  /**
   * Returns a boolean read from the stream or throws the CarobException
   * that came instead.
   * 
   * @throws SocketIOException stream or protocol error
   * @throws CarobException coming from the controller
   */
  bool                receiveBoolOrException() throw (SocketIOException,
                                                      BackendException,
                                                      ControllerException,
                                                      ProtocolException,
                                                      UnexpectedException);
  /**
   * Returns a int read from the stream or throws the CarobException that
   * came instead.
   * 
   * @throws IOException stream or protocol error
   * @throws CarobException coming from the controller
   */
  int32_t             receiveIntOrException() throw (SocketIOException,
                                                     BackendException,
                                                     ControllerException,
                                                     ProtocolException,
                                                     UnexpectedException);
  /**
   * Returns a long read from the stream or throws the CarobException
   * that came instead.
   * 
   * @throws IOException stream or protocol error
   * @throws CarobException coming from the controller
   */
  int64_t             receiveLongOrException() throw (SocketIOException,
                                                      BackendException,
                                                      ControllerException,
                                                      ProtocolException,
                                                      UnexpectedException);
  /**
   * Returns a string read from the stream or throws the CarobException
   * that came instead.
   * 
   * @throws SocketIOException stream or protocol error
   * @throws CarobException coming from the controller
   */
  std::wstring        receiveStringOrException() throw (SocketIOException,
                                                        BackendException,
                                                        ControllerException,
                                                        ProtocolException,
                                                        UnexpectedException);
  /**
   * Deserialize SQL warnings from the stream: converts BackendDriverException
   * to an SQLWarning chain
   * 
   * @return the deserialized warning chain
   * @throws IOException stream error
   * @throws ProtocolException protocol error
   * @throws SerializableException
   */
  SQLWarning*         receiveSQLWarnings() throw (SocketIOException,
                                                  BackendException,
                                                  ControllerException,
                                                  ProtocolException,
                                                  UnexpectedException);

  /**
   * Deserialize an exception from the stream: converts explicit protocol
   * typing into java types.
   * 
   * @return the deserialized exception read from the stream
   * @throws IOException in case of stream error, the deserialized exception
   *         otherwise
   */
  void                receiveException() throw (SocketIOException,
                                                BackendException,
                                                ControllerException,
                                                ProtocolException,
                                                UnexpectedException);
  /**
   * Returns a DriverResultSet read from the stream or throws the Exception that
   * came instead
   * 
   * @throws SocketIOException stream or protocol error
   * @throws CarobException received from the controller
   */
  DriverResultSet*    receiveResultSet() throw (SocketIOException,
                          BackendException, ControllerException,
                          ProtocolException, NotImplementedException,
                          UnexpectedException);
  /**
   * Check if the given commit has been successfully performed.
   *
   * @return transaction id if the commit has been successfully performed
   */
  int64_t             retrieveCommitResult() throw (SocketIOException,
                          BackendException, ControllerException,
                          ProtocolException, UnexpectedException);
  /**
   * Check if the given rollback has been successfully performed.
   * 
   * @return the transaction id if the rollback has been successfully performed
   */
  int64_t             retrieveRollbackResult() throw (SocketIOException,
                          BackendException, ControllerException,
                          ProtocolException, UnexpectedException);
  /**
   * Check if the given query already executed or not on the controller we are
   * currently connected to.
   * 
   * @param request the request to check
   * @return int the number of updated rows or -1 if not found
   */
  ResultAndWarnings   retrieveExecuteUpdateResult(const Request &request)
                          throw (SocketIOException, BackendException,
                          ControllerException, ProtocolException,
                          UnexpectedException);
  /**
   * Check if the given query already executed or not on the controller we are
   * currently connected to. If so, return this query previous result
   * 
   * @param request the request to check
   * @return a <code>List</code> of ResultSetOrUpdateCount
   */
  ResultAndWarnings   retrieveExecuteResult(const Request &request)
                          throw (SocketIOException, BackendException,
                          ControllerException, ProtocolException,
                          UnexpectedException);
  /**
   * Check if the given query already executed or not on the controller we are
   * currently connected to. If so, return this query previous result
   * 
   * @param request the request to check
   * @return previous result of the given query, 
   * @throws DriverSQLException if an error occurs
   */
  ResultAndWarnings   retrieveExecuteUpdateWithKeysResult(
                          const Request &request) throw (SocketIOException,
                          BackendException, ControllerException,
                          ProtocolException, UnexpectedException);
  /**
   * Fetches multiple results from a query executed with Statement.execute()
   * @return the list of results
   */
  std::list<ResultSetOrUpdateCount> fetchMultipleResultsFromStream()
                          throw (SocketIOException, BackendException,
                          ControllerException, ProtocolException,
                          UnexpectedException);
  /**
   * Tries first to reconnect to the same controller. If unsuccessfull, notifies
   * the suspicion of failure for the current controller and reconnects to a
   * controller chosen using the policy specified in the connection parameters
   * of this connection.
   * 
   * @throw ConnectionException if the next controller (get from connection
   *        policy) is also down
   * @throw NoMoreControllerException if no valid controller left
   * @throw DriverException if controllers are misconfigurated
   */
  void                tryToReconnect() throw (NoMoreControllerException,
                            DriverException, UnexpectedException);
  /**
   * Calls reconnect until :
   *  - success
   *  - NoMoreControllerException
   *  - Other unrecoverable error
   * 
   * @throw ConnectionException if no valid controller is left
   * @throw DriverException if controllers are misconfigurated
   */
  void                reconnectOrDie() throw (ConnectionException,
                            DriverException, UnexpectedException);
  /**
   * Forbids Connection creation that would lead to unexpected state.
   * A connection is either created, ie. connected to a controller or destroyed
   * @see #Connection(const ConnectionParameters&)
   */
  Connection();
};
} //namespace CarobNS
#endif //_CONNECTION_H_
