/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert, Zsolt Simon
 */

#ifndef CONTROLLERPINGER_HPP_
#define CONTROLLERPINGER_HPP_

namespace CarobNS
{
  
class ControllerInfo;

/**
 * Singleton (library-wide) class that manages the list of controllers suspected
 * of failure.<br>
 * Used by controller pools, provides 3 main functions: one for adding a suspect
 * in the list, one for knowing if a controller is already suspected, and on to
 * update the list (ping the controllers to check if they are up again
 */
class ControllerPinger
{
public:
  /**
   * Singleton-access function: retrieves the existing or newly created instance
   * of the ControllerPinger
   */ 
  static ControllerPinger&    getInstance();
  /**
   * Adds the given controller to the list of suspects.<br>
   * The given controllerInfo argument will be *copied* in the suspected
   * controllers list, but this function will modify the given controller info
   * by setting its failure_number field
   * @param controllerInfoPrm the controller suspected of failure
   */
  void                        suspectControllerOfFailure(
                                  ControllerInfo& controllerInfoPrm);
  /**
   * Returns true if the specified controller is suspected of failure.<br>
   * @param controllerInfoPrm the controller to check
   * @return true if the controller is in the suspect list
   */
  bool                        isSuspectedOfFailure(
                                  const ControllerInfo& controllerInfoPrm);
  /**
   * Tries to connect to each suspected controller with a non-blocking connect
   * and checks the connection state. In case of connection success, send the
   * ping command and remove the controller from the suspect list.
   */
  void                        run();
  /**
   * Gives the number of controller failures since this process started
   */
  static int                  getNumberOfFailures();
private:
  // no suspected_controllers list here. As this vector is library-wide, it is
  // declared in an anymous namespace in the .cpp
  
  /** Unique instance of this class */
  static ControllerPinger*    unique_instance_ptr;
  /**
   * Number of milliseconds to wait for the controllers to respond to a ping
   * connection attempt
   */
  static const int            poll_timeout_ms = 20;
  /** Prevent creation from the outside */
  ControllerPinger() {};
  /** Prevent destruction from the outside */
  virtual ~ControllerPinger() {};
  /**
   * Removes the specified controller from the list of suspected controllers.
   * If the given controller is not in the list, logs a warning and returns
   * @param controller the controller to remove from the list
   */
  static void                 removeControllerFromSuspectList(
                                  const ControllerInfo& controller);
};

} //namespace CarobNS

#endif /*CONTROLLERPINGER_HPP_*/
