/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert, Zsolt Simon
 */
#ifndef CONTROLLERPOOL_H_
#define CONTROLLERPOOL_H_

#include "CarobException.hpp"
#include "CriticalSection.hpp"

#include <vector>
#include <map>

namespace CarobNS
{

enum ConnectPolicy
{
  ROUND_ROBIN
};


/**
 * Exception on connection send when no more controller is reachable in the
 * controller list
 */
class NoMoreControllerException : public DriverException
{
public:
  /**
   * Constructs a NoMoreControllerException with the given message
   * @param s error message of the exception
   */
  NoMoreControllerException(std::wstring s) : DriverException(s) {};
};

class ControllerInfo;
/**
 * Abstract class over each policy implementation, which is used by the driver
 * to choose a controller to connect to.
 */
class AbstractControllerPool
{
public:
  /**
   * Creates a new <code>AbstractControllerPool</code> object
   * 
   * @param controllerList the controller list on which the policy applies
   * @param retryIntervalInMs IGNORED ! interval in milliseconds before retrying
   *          to re-connect to a controller that has failed
   */
  AbstractControllerPool(const std::vector<ControllerInfo>& controllerList,
      long retryIntervalInMs = 0) throw (DriverException, UnexpectedException);
  /**
   * Clears list of controllers
   */
  virtual ~AbstractControllerPool();
  /**
   * Gets a controller using the implementation specific policy
   * @return <code>ControllerInfo</code> of the selected controller
   * @throws NoMoreControllerException if no controller in the controller list
   *         is reachable
   */
  virtual ControllerInfo      getController() throw (NoMoreControllerException,
      UnexpectedException) = 0;
  /**
   * Tells the pool that a new connection is using it
   */
  void                        addRef() { ref_counter++; }
  /**
   * Tells the pool that one of its 'user' (ie. connection) doesn't use it
   * anymore, and return the new number of references
   */
  int                         removeRef() { ref_counter--; return ref_counter; }
  /**
   * Display useful info in a string
   */
  operator std::wstring() const;
protected:
  /** list of valid controllers */
  std::vector<ControllerInfo> controller_list;
  /** class-wide mutex */
  CriticalSection             pool_CS;
private:
  /** references counter managed by controller pool manager */
  int                         ref_counter;
};

/**
 * Implements a controller selection policy that iterates through the controller
 * list, one after the other, starting with the first one in the list
 */
class RoundRobinControllerPool : public AbstractControllerPool
{
public:
  /**
   * Creates a new <code>RoundRobinControllerPool</code> object
   * 
   * @param controllerList the controller list on which the policy applies
   * @param retryIntervalInMs IGNORED ! interval in milliseconds before retrying
   *          to re-connect to a controller that has failed
   */
  RoundRobinControllerPool(const std::vector<ControllerInfo>& controllerList,
    long retryIntervalInMs = 0) throw (DriverException, UnexpectedException);
  /**
   * @see AbstractControllerPool#getController()
   */
  ControllerInfo              getController() throw (NoMoreControllerException,
                                  UnexpectedException);
private:
  /** index of last controller given by getController */ 
  int                         index;
};

/**
 * Utility struct for controller list / controller pool association (see
 * controller pool manager
 */
typedef struct
{
  /** List of controllers (warning: order matters !) */
  std::vector<ControllerInfo> controllers;
  /** Associated Connection policy */
  ConnectPolicy               policy;
} PoolIndex;

/**
 * PoolIndex "Less Than" functor.<br>
 * First compares policies (just by comparing their enum values)<br>
 * Then the number of controllers<br>
 * Finally the controllers themselves using ControllerInfo '<' operator<br>
 * This functor is used by ControllerPoolManager for sorting values in the pool
 * map
 */
struct ltPoolIndex
{
  /**
   * Less than operator
   */
  bool operator()(const PoolIndex& pi1, const PoolIndex& pi2) const;
};

/**
 * Holds a list of controller pools, that are indexed by their policy and list 
 * of controllers. Permits to have only one instance of ControllerPool for the
 * same controllers and policy, thus to share the connection info (ie. on which
 * controller the last connection occured).<br>
 */
class ControllerPoolManager
{
public:
  /**
   * Returns a pool for the given pair controllers/policy.<br>
   * If the controller pool list already contains a pool suitable for given
   * parameters, returns it after incrementing the reference count on it. If not
   * found, creates a new controller pool, adds it to the list, increments its
   * reference counter and returns it.
   * @param ctrls (ordered) controller list to connect to
   * @param cp policy to use
   * @return a pool following the given policy for the given controllers
   * distribution
   * @throws DriverException if the given policy is invalid 
   */
  static AbstractControllerPool&  getPool(std::vector<ControllerInfo> ctrls,
                                     ConnectPolicy cp)
                                     throw (DriverException, UnexpectedException);
  /**
   * Decrements reference counter on the given pool. If no more references are
   * made on this pool, removes it from the pool list and deletes it
   * @param pool the pool to release
   */
  static void                     freePool(AbstractControllerPool* pool);
protected:
private:
  /** To manage pool map concurrent accesses */
  static CriticalSection          pool_map_CS;
  /** The map that associates controller lists to pools */
  static std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>  pool_map;
};

} //namespace CarobNS

#endif /*CONTROLLERPOOL_H_*/
