/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon
 */

#ifndef SOCKET_H_
#define SOCKET_H_

#include "StringCodecs.hpp"
#include "CarobException.hpp"

#include "Common.hpp" // for java_byte: cannot forward declare typedefs

#include <string>

#ifdef __MINGW32__
  #include <winsock2.h>
#else
  #include <netinet/in.h> //for in_addr_t
#endif


namespace CarobNS {

/**
 * Generic socket wrapper class to exchange data with Java world.
 * This class can read and write basic Java types over a socket. It 
 * does *not* follow any protocol but java pure serialization.
 * @see DriverSocket class for sequoia protocol.
 */
class JavaSocket
{
public:
  /**
   * Default constructor - No connection here, just creates and empty socket
   */
  JavaSocket();
  /**
   * Destructor for disconnection - closes the socket.
   * @throws SocketIOException
   */
  virtual ~JavaSocket() throw (SocketIOException, UnexpectedException);
  /**
   * Gives socket validity status.
   * @return true if the socket is valid (has been created), so
   * {@link #connectTo(const std::wstring& host, in_port_t port)} can be called
   */
  inline bool   isValid() const { return socketFd != -1; }
  /**
   * Gives connection status
   * @return true if the socket is connected to a host
   */
  inline bool   isConnected() const { return connected; }

  /**
   * Creates the socket.
   * @param blocking whether or not socket operations must be blocking (default
   *        is true, blocking socket)
   * @return true upon successfull creation, false otherwise
   * @throws ConnectionException
   */
  bool          create(bool blocking = true)
                    throw (ConnectionException, UnexpectedException);
  /**
   * Connects to the given host/port
   * @param host  the host to connect to as a string. Can be either numeric
   *              IP or fully qualified hostname
   * @param port port as a in_addr_t (unsigned int)
   * @return true upon successfull connection, false otherwise
   * @throws ConnectionException
   */
  bool          connectTo(const std::wstring& host, in_port_t port)
                    throw (ConnectionException, UnexpectedException);
  /** 
   * Closes connection.
   * @return true if the socket was successfully closed
   * @throws SocketIOException
   */
  virtual bool  closeSocket() throw (SocketIOException, UnexpectedException);
  
  /**
   * Writes a UTF-8 encoded wstring to the socket. This is NOT Java's
   * modified UTF-8, see SEQUOIA-133.
   * @param str string to write
   * @return the number of bytes send
   * @throws SocketIOException
   */
  size_t        writeJavaUTF(const std::wstring& str) const
    throw (SocketIOException, CodecException, UnexpectedException);
  /**
   * Reads UTF-8 encoded string from socket. This is NOT Java's
   * modified UTF-8, see SEQUOIA-133.
   * @param str string to read
   * @return the length read in bytes
   * @throws SocketIOException
   */
  size_t        readJavaUTF(std::wstring& str) const
    throw (SocketIOException, CodecException, UnexpectedException);
  /**
   * Writes 32bits-integer to socket
   * @param i integer to send
   * @throws SocketIOException
   */
  void          writeJavaInt(int32_t i) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Reads 32bits-integer from socket
   * @param i resulting integer read from stream
   * @throws SocketIOException
   */
  void          readJavaInt(int32_t& i) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Writes 64bits-long to socket
   * @param l long to send
   * @throws SocketIOException
   */
  void          writeJavaLong(int64_t l) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Reads 64bits-long from socket
   * @param l resulting long read from stream
   * @throws SocketIOException
   */
  void          readJavaLong(int64_t& l) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Writes a boolean (as a byte) to the socket
   * @param b the boolean to send
   * @throws SocketIOException
   */
  void          writeJavaBool(bool b) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Reads a boolean (byte) from the socket
   * @param b resulting boolean to read from stream
   * @throws SocketIOException
   */
  void          readJavaBool(bool& b) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Reads an array of bytes from the socket. The array must be allocated !
   * @param length number of bytes to read
   * @param data allocated array of (at least) length bytes
   * @throws SocketIOException
   */
  void          readJavaBytes(int32_t length, java_byte* data) const
                    throw (SocketIOException, UnexpectedException);
  /**
   * Writes an array of bytes to the socket.
   * @param length number of bytes to write
   * @param data array of (at least) length bytes
   * @throws SocketIOException
   */
  void          writeJavaBytes(int32_t length, java_byte* data) const
                    throw (SocketIOException, UnexpectedException);
  /**
   * Gets the socket file descriptor
   * @return the file descriptor
   */
  int           getFd() const { return socketFd; }
protected:
  /**
   * Substitute for recv. Loops/blocks until full length has been received.
   * @param buf data to send
   * @param len full buffer length
   * @param flags send options, see recv man page
   * @return the total number of bytes send, -1 in case of failure
   */
  int32_t       recvFully(void *buf, const int len, const int flags)
                    const;
  /**
   * Wrapper over send(...) function to handle errors and throw exceptions
   * @param fctName name of the calling function (for logging purposes)
   * @param objName name of the object to be send (for logging purposes)
   * @param buf data to be send
   * @param len length of buf
   * @param flags send option, see recv man page
   * @throws SocketIOException
   */
  void          sendToSocket(const std::wstring& fctName,
                    const std::wstring& objName, const void* buf, int len,
                    int flags) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Wrapper around recv(...) function to handle errors and throw exceptions
   * @param fctName name of the calling function (for logging purposes)
   * @param objName name of the object to be send (for logging purposes)
   * @param buf buffer in which to put the received data in
   * @param len length of buf
   * @param flags send option, see send man page
   * @throws SocketIOException in case of error
   */
  void          receiveFromSocket(const std::wstring& fctName,
                    const std::wstring& objName, 
                    void *buf, int len, int flags)
                    const throw (SocketIOException, UnexpectedException);

private:
  /** Socket file descriptor */
  int           socketFd;
  /** true if the socket is connected to a host */
  bool          connected;

  /**
   * Function converts the unsigned 64bit integer from network byte order to host byte order.
   * @param n number to be converted
   * @return the converted number
   */
  const uint64_t ntohll(const uint64_t &n) const;

  /**
   * Function converts the unsigned 64bit integer from host byte order to network byte order.
   * @param n number to be converted
   * @return the converted number
   */
  const uint64_t htonll(const uint64_t &n) const;
};

} //namespace CarobNS
#endif /*SOCKET_H_*/
