/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert
 */

#include "CarobException.hpp"

#include "Common.hpp"
#include "DriverSocket.hpp"

#include <sstream>

#ifdef __GLIBC__  // from features.h
#define GLIB_BACKTRACE
#include <execinfo.h>
#include <stdlib.h>
#endif // __GLIBC__

using std::wstring;

using namespace CarobNS;

/** constructs the BackTrace of the current place */
BackTrace::BackTrace()
{
#ifdef GLIB_BACKTRACE
  void * stack_pointers[100];
  size_t size = backtrace (stack_pointers /* set it */, 100);
  char **cstrings = backtrace_symbols (stack_pointers /* use it */, size);
  for (size_t i=0; i<size; i++)
    this->push_back(cstrings[i]);
  free(cstrings);
#endif
}


CarobException::~CarobException()
{
  if (causePtr != NULL)
  {
    delete causePtr;
    causePtr = NULL;
  }
}

/** Copy constructor */
CarobException::CarobException(const CarobException& ce) :
  message(ce.message),
  stackTrace(ce.stackTrace),
  SQLState(ce.SQLState),
  errorCode(ce.errorCode),
  backTrace(ce.backTrace)
{
  if (ce.causePtr != NULL)
  { // copy chain by recursively calling copy constructor
    causePtr = new CarobException(*(ce.causePtr));
  }
  else
  {
    causePtr = NULL;
  }
}

CarobException::CarobException(const DriverSocket& sock)
{
  bool hasNextException;

  sock>>message;
  sock>>hasNextException;
  if (hasNextException)
    causePtr = new CarobException(sock);
  else
    causePtr = NULL;
  
  // receive stackTrace
  int stLength;
  sock>>stLength;
  for (int i = 0; i < stLength; i++)
  {
    stackTrace.push_back(StackTraceElement(sock));
  }
  sock>>SQLState;
  sock>>errorCode;
}
void UnexpectedException::convertUnexpected()
{
  if (isErrorEnabled())
  {
    logError(L"convertUnexpectedException",
             L"Unexpected exception caught... Throwing Carob UnexpectedException");
  }
  throw UnexpectedException(L"Error not intercepted"); 
}

StackTraceElement::StackTraceElement(const DriverSocket& ds)
    throw (SocketIOException, UnexpectedException)
{
  ds>>declaringClass;
  ds>>methodName;
  ds>>fileName;
  ds>>lineNumber;
}

std::wstring
StackTraceElement::toWString() const
{ 
    std::wostringstream oss;
    oss << L"at " << declaringClass << L"." << methodName
        << L"(" << fileName << L":" << lineNumber << L")";
    return oss.str();
}


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
