/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert, Zsolt Simon
 */

#include "ControllerPool.hpp"

#include "ControllerPinger.hpp"
#include "DriverSocket.hpp"
#include "Connection.hpp"
#include "Statement.hpp"
#include "ParameterStatement.hpp"
#include "TypeTag.hpp"
#include "Request.hpp"
#include "RequestWithResultSetParameters.hpp"

#include "ConnectionParameters.hpp"
#include "CarobException.hpp"

#include <string>

#ifdef __MINGW32__
  #define sleep _sleep
#else
  #include <unistd.h> // sleep()
#endif

using std::wstring;

using namespace CarobNS;

#define RECONNECT_RETRIES 10
#define CONNECT_RETRIES 3

// The reconnection macro:
// When a SocketIOException occurs between these to macros, sleep and try to
// reconnect(). Sleeping time depends on the retry #. Only N retries are
// performed
#define FO_TRY_NTIMES(N)\
for (int foCnt=0; foCnt<N; foCnt++)\
{\
  try\
  {\
    /* Code to monitor will go here */
#define FO_CATCH_NTIMES\
    break; /* if no exception, get out of the for */\
  }\
  catch (SocketIOException sioe)\
  {\
    ::sleep(static_cast<int>(foCnt*0.5)); /* sleep incrementally */\
    reconnectOrDie();\
  }\
}


Connection::Connection(ConnectionParameters& prms)
    throw (DriverException, ConnectionException, AuthenticationException, UnexpectedException) :
  driverSocketPtr(NULL),
  isClosed(true),
  autoCommit(true),
  readOnly(false),
  writeExecutedInTransaction(false),
  mustBeginTransaction(false),
  persistent_connection(false),
  retrieve_sql_warnings(false),
  first_warning(NULL),
  controller_pool(prms.getControllerPool()),
  parameters(prms)
{
  wstring fctName(L"Connection::Connection");
  persistent_connection = parameters.getPersistentConnection();
  retrieve_sql_warnings = parameters.getRetrieveSQLWarnings();
  for (int connectRetry = 0; connectRetry<CONNECT_RETRIES; connectRetry++)
  {
    try
    {
      connected_controller = controller_pool.getController();
  
      // Do the authentication stuff
      initConnection();
      // receive ack and other params
      finalizeConnect();
      isClosed = false;
      return;
    }
    // SocketIO and connection exceptions indicate dead controller => try to
    // connect to the next one in the controller list
    catch (SocketIOException sioe)
    {
      //Do some clean up
      delete driverSocketPtr; driverSocketPtr = NULL;
      //suspect failing controller and try next one
      ControllerPinger::getInstance().suspectControllerOfFailure(connected_controller);
    }
    catch (ConnectionException ce)
    {
      //Do some clean up
      delete driverSocketPtr; driverSocketPtr = NULL;
      //suspect failing controller and try next one
      ControllerPinger::getInstance().suspectControllerOfFailure(connected_controller);
    }
    catch (NoMoreControllerException nmce)
    {
      // Clean all: we are in constructor, destructor won't be called !
      delete driverSocketPtr; driverSocketPtr = NULL;
      parameters.releaseControllerPool();
      throw (ConnectionException(nmce.description()));
    }
    // Other exceptions must be reported to the user
    catch (...)
    {
      // Clean all: we are in constructor, destructor won't be called !
      delete driverSocketPtr; driverSocketPtr = NULL;
      parameters.releaseControllerPool();
      throw;
    }
  }
}

Connection::~Connection()
{
  close();
  delete driverSocketPtr;
  parameters.releaseControllerPool();
}

void Connection::sendCommand(const DriverSocket& socket, int command)
    throw (SocketIOException, ProtocolException, UnexpectedException)
{
  int ack = -1;
  socket>>ack;
  if (ack != ControllerPrompt)
    throw ProtocolException(
        L"Protocol corruption while trying to send command: "
        + toUserString(command) + L". Check the previous command");
  socket<<CommandPrefix;
  socket<<static_cast<int32_t>(command);
}

void Connection::initConnection()
    throw (DriverException, ConnectionException, AuthenticationException,
        UnexpectedException)
{
  wstring fctName(L"Connection::initConnection");
  bool  protocolVersionSend = false,
        dbNameSend = false,
        userNameSend = false,
        userPassSend = false;
  if (isInfoEnabled())
    logInfo(fctName, L"Authenticating with controller "
                      + connected_controller.getHostName() + L":"
                      + toUserString(static_cast<int>(connected_controller.getHostPort())));
  try
  {
    //Here is the connection protocol...
    // 1. connect to the controller
    driverSocketPtr = new DriverSocket(connected_controller.getHostName(),
                                       connected_controller.getHostPort());
    // 2. sent the Protocol version information
    *driverSocketPtr<<ProtocolVersion;
    protocolVersionSend = true;
    // 3. send the database name
    *driverSocketPtr<<parameters.getDatabaseName();
    dbNameSend = true;
    // 4. send user/pass
    *driverSocketPtr<<parameters.getUserName();
    userNameSend = true;
    *driverSocketPtr<<parameters.getUserPass();
    userPassSend = true;
  }
  catch (ConnectionException connExcpt)
  {
    wstring msg = L"Unable to connect to controller on "
                  + connected_controller.getHostName()
                  + L":" + toUserString(static_cast<int>(connected_controller.getHostPort()))
                  + L" (" + connExcpt.description() + L").";
    if (isErrorEnabled())
      logError(fctName, msg);
    throw ConnectionException(msg);
  }
  catch (SocketIOException sockIOExcpt)
  {
    wstring msg = L"Could not authentify to " + connected_controller.getHostName()
                  + L":" + toUserString(static_cast<int>(connected_controller.getHostPort()));
    if (!protocolVersionSend)
    {
      msg+=L". Error while sending protocol version ("
           +sockIOExcpt.description()+L")";
      if (isErrorEnabled())
      {
        logError(fctName, msg);
      }
      throw AuthenticationException(msg);
    }
    else if (!dbNameSend)
    {
      msg+=L". Error while sending database name ("
           +sockIOExcpt.description()+L")";
      if (isErrorEnabled())
      {
        logError(fctName, msg);
      }
      throw AuthenticationException(msg);
    }
    else if (!userNameSend)
    {
      msg+=L". Error while sending user name ("
           +sockIOExcpt.description()+L")";
      if (isErrorEnabled())
      {
        logError(fctName, msg);
      }
      throw AuthenticationException(msg);
    }
    else if (!userPassSend)
    {
      msg+=L". Error while sending user password ("
           +sockIOExcpt.description()+L")";
      if (isErrorEnabled())
      {
        logError(fctName, msg);
      }
      throw AuthenticationException(msg);
    }
    else
    {
      msg+=L". Unknown error:"
           +sockIOExcpt.description();
      if (isErrorEnabled())
      {
        logError(fctName, msg);
      }
      throw AuthenticationException(msg);
    }
  }
}

void Connection::finalizeConnect() throw (ConnectionException, AuthenticationException,
    UnexpectedException)
{
  wstring fctName(L"Connection::finalizeConnect");
  bool vdbFound = false,
       vdbFoundRead = false,
       authenticated = false,
       authenticatedRead = false,
       lineSeparatorSent = false,
       persistentConnectionSent = false,
       persistentConnectionRead = false;
  try
  {
    *driverSocketPtr >> vdbFound;
    vdbFoundRead = true;
    if (!vdbFound)
    {
      wstring vdbReason;
      *driverSocketPtr >> vdbReason;
      throw AuthenticationException(L"Connection failed: "+vdbReason);
    }
    *driverSocketPtr >> authenticated;
    authenticatedRead = true;
    if (!authenticated)
    {
      wstring authReason;
      *driverSocketPtr >> authReason;
      throw AuthenticationException(L"Authentication failed: "+authReason);
    }
    // So the controller can correctly parse our requests
    *driverSocketPtr << LINE_SEPARATOR;
    lineSeparatorSent = true;
    *driverSocketPtr << persistent_connection;
    persistentConnectionSent = true;
    if (persistent_connection)
    {
      bool success = false;
      *driverSocketPtr >> success;
      if (success)
        *driverSocketPtr >> persistent_connection_id;
      else
        throw AuthenticationException(fctName + 
            L" No more connection available in the cluster");
    }
    persistentConnectionRead = true;
    *driverSocketPtr << retrieve_sql_warnings;
  }
  catch (SocketIOException sockIOExcpt)
  {
    wstring msg(fctName + L" Authentication failed. Error while ");
    if (!authenticatedRead)
      msg += L"reading controller authenticated ack";
    else if (!lineSeparatorSent)
      msg += L"sending line separator";
    else if (!persistentConnectionSent)
      msg += L"sending whether to use persistent connections";
    else if (!persistentConnectionRead)
      msg += L"reading persistent connection id";
    else
      msg += L"sending whether to retrieve SQL warnings";
    throw AuthenticationException(msg
                                  + L"(" + sockIOExcpt.description() + L")");
  }
}

void Connection::deleteStatement(Statement* stPtr) throw (DriverException, UnexpectedException)
{
  if (stPtr != NULL) // if null, we don't do anything
  {
    for (std::list<Statement*>::iterator iter = created_statements.begin();
        iter != created_statements.end(); iter++)
    {
      if (*iter == stPtr)
      {
        created_statements.erase(iter);
        delete stPtr;
        return;
      }
    }
    // We shouldn't come here. It means we have been asked to delete a statement
    // we don't own
    throw DriverException(L"deleteStatement: the given statement is not owned by this connection!");
  }
}

bool Connection::close()
{
  wstring fctName(L"Connection::close");

  //Wait until other methods/Commands are done
  LockScope ls(&connectionCS);

  isClosed = true;

  if (isDebugEnabled())
    logDebug(fctName, L"Deleting statements created by this connection");
  //delete all statement created by this connection
  for (std::list<Statement*>::iterator iter = created_statements.begin();
      iter != created_statements.end(); iter++)
  {
    delete *iter;
  }
  created_statements.clear();

  return closeSocket();
}

bool Connection::closeSocket()
{
  wstring fctName(L"Connection::closeSocket");
  bool closeOK = false;

  if (driverSocketPtr == NULL)
  {
    return false;
  }

  if (isDebugEnabled())
    logDebug(fctName, L"Closing socket to controller");
  try
  {
    // Try to cleanly close the connection (also on controller side)
    sendCommand(*driverSocketPtr, Close);
    closeOK = receiveBoolOrException();
  }
  catch (...)
  {
    if (isWarnEnabled())
      logWarn(fctName, L"Could not inform controller we are closing connection");
  }
  try
  {
    //now close the socket itself
    closeOK &= driverSocketPtr->closeSocket();
    return closeOK;
  }
  //catch any exception, log an error and return false (no exception forward)
  catch (...)
  {
    if (isWarnEnabled())
    {
      logWarn(fctName, L"Caught exception while closing connection");
    }
    //Don't re-throw = transparent closing
  }
  return false;
}

void Connection::setAutoCommit(bool autoCommitPrm)
    throw (SocketIOException, BackendException, ControllerException, 
    ProtocolException, ConnectionException, UnexpectedException)
{
  wstring fctName(L"Connection::setAutoCommit");

  LockScope ls(&connectionCS);
  
  if (mustBeginTransaction) // so this.autoCommit is false
  {
    if (autoCommitPrm)
    { // Just cancel the need to start a transaction
      autoCommit = true;
      mustBeginTransaction = false;
      return;
    }
  }
    
  // nothing to change
  if (autoCommit == autoCommitPrm)
  {
    return;
  }
  // autocommit false -> true
  if (autoCommitPrm)
  {
    if (isDebugEnabled())
      logDebug(fctName, L"Setting connection in autocommit mode");

    doCommit();
    autoCommit = true;
    return;
  }
  else // autocommit true -> false
  {
    
    mustBeginTransaction = true;
    autoCommit = false;

    if (isDebugEnabled())
      logDebug(fctName, L"Transaction" + toUserString(transactionId)
                        + L" has been started");
  }
}

void Connection::beginTransactionIfNeeded() throw (SocketIOException,
    BackendException, ControllerException, ProtocolException,
    ConnectionException, UnexpectedException)
{
  if (mustBeginTransaction)
  {
    FO_TRY_NTIMES(RECONNECT_RETRIES)
      sendCommand(*driverSocketPtr, Begin);
      transactionId = receiveLongOrException();
      mustBeginTransaction = false;
    FO_CATCH_NTIMES
  }
}

void Connection::setReadOnly(bool readOnlyPrm) throw (SocketIOException,
DriverException, UnexpectedException)
{
  checkIfConnected();
  FO_TRY_NTIMES(RECONNECT_RETRIES)
    sendCommand(*driverSocketPtr, SetReadOnly);
    *driverSocketPtr << readOnlyPrm;
    receiveBoolOrException();
    readOnly = readOnlyPrm;
  FO_CATCH_NTIMES
}

void Connection::commit() throw (SocketIOException, DriverException,
    ProtocolException, ConnectionException, UnexpectedException)
{
  wstring fctName(L"Connection::commit");
  
  LockScope ls(&connectionCS);
  checkIfConnected();

  if (autoCommit)
  {
    throw DriverException(L"Trying to commit a connection in autocommit mode");
  }

  // Check if we are not committing an empty transaction (not started yet)
  if (mustBeginTransaction)
    return;

  doCommit();
}

void Connection::doCommit()  throw (SocketIOException, DriverException,
    ProtocolException, ConnectionException, UnexpectedException)
{
  try
  {
    sendCommand(*driverSocketPtr, Commit);
    int64_t acknowledgedTransactionId = receiveLongOrException();
    if (acknowledgedTransactionId != transactionId)
    {
      throw (ProtocolException(
          L"Protocol error during commit (acknowledge transaction ID = "
          + toUserString(acknowledgedTransactionId)
          + L", expected transaction ID = "
          + toUserString(transactionId) + L")"));
    }
  }
  catch (SocketIOException e)
  { // Connection failed, try to reconnect and re-exec the commit
    reconnectOrDie();

    // FIXME: we should probably retry NTIMES here
    int64_t acknowledgedTransactionId = retrieveCommitResult();
    if (acknowledgedTransactionId != transactionId)
    {
      throw (DriverException(
          L"Error during commit failover (acknowledge transaction ID = "
          + toUserString(acknowledgedTransactionId)
          + L", expected transaction ID = " + toUserString(transactionId) + L")"));
    }

    // The controller will automatically redo the commit if it was not done
    // earlier so we can safely return here, this is a success.
  }

  // Success (with or without reconnection)

  // Commit must be followed by a BEGIN
  mustBeginTransaction = true;
  writeExecutedInTransaction = false;

}

void Connection::rollback() throw (SocketIOException, DriverException,
    ProtocolException, ConnectionException, UnexpectedException)
{
  wstring fctName(L"Connection::rollback");
  
  LockScope ls(&connectionCS);

  checkIfConnected();

  if (autoCommit)
  {
    throw DriverException(L"Trying to rollback a connection in autocommit mode");
  }

  // Check if we are not rollbacking an empty transaction (not started yet)
  if (mustBeginTransaction)
    return;

  try
  {
    sendCommand(*driverSocketPtr, Rollback);
    int64_t acknowledgedTransactionId = receiveLongOrException();
    if (acknowledgedTransactionId != transactionId)
    {
      throw (ProtocolException(
          L"Protocol error during rollback (acknowledge transaction ID = "
            + toUserString(acknowledgedTransactionId)
            + L", expected transaction ID = "
            + toUserString(transactionId) + L")"));
    }
    // Rollback is followed by a BEGIN
    mustBeginTransaction = true;
  }
  catch (SocketIOException e)
  { // Connection failed, try to reconnect and re-exec the rollback
    reconnectOrDie();
    int64_t acknowledgedTransactionId = retrieveRollbackResult();
    if (acknowledgedTransactionId != transactionId)
    {
      throw (DriverException(
          L"Error during rollback failover (acknowledge transaction ID = "
          + toUserString(acknowledgedTransactionId)
          + L", expected transaction ID = " + toUserString(transactionId) + L")"));
    }
    // The controller will automatically redo the rollback if it was not
    // done earlier so we can safely return here, this is a success.
    return;
  }
}

Statement* Connection::createStatement() throw (SocketIOException,
    UnexpectedException)
{
  checkIfConnected();
  Statement* stPtr = new Statement(this);
  created_statements.push_back(stPtr);
  return stPtr;
}

ParameterStatement* Connection::createParameterStatement(const std::wstring &query) 
    throw (SocketIOException, UnexpectedException)
{
  checkIfConnected();
  ParameterStatement* pstPtr = new ParameterStatement(this, query);
  created_statements.push_back(pstPtr);
  return pstPtr;
}


void Connection::checkIfConnected() const throw (SocketIOException,
    UnexpectedException)
{
  if (driverSocketPtr == NULL || !driverSocketPtr->isConnected())
    throw SocketIOException(
        L"Request cannot be processed : connection is closed!");
}

DriverResultSet* Connection::preparedStatementGetMetaData(const wstring &sqlTemplate)
    throw (SocketIOException, BackendException, ControllerException, 
    ProtocolException, NotImplementedException, ConnectionException,
    UnexpectedException)
{
  checkIfConnected();
  wstring fctName(L"Connection::PreparedStatementGetMetaData");

  LockScope ls(&connectionCS);

  FO_TRY_NTIMES(RECONNECT_RETRIES)
    sendCommand(*driverSocketPtr, PreparedStatementGetMetaData);
    *driverSocketPtr << sqlTemplate;
  
    if (isDebugEnabled())
      logDebug(fctName, L"Statement sent successfully");
  
    return receiveResultSet();
  FO_CATCH_NTIMES

  return NULL; //avoid compiler warning
}

void Connection::setConnectionParametersOnRequest(Request &request)
{
  request.setIsAutoCommit(autoCommit);
}

DriverResultSet* Connection::statementExecuteQuery
    (RequestWithResultSetParameters& request)
    throw (SocketIOException, BackendException, ControllerException,
    ProtocolException, NotImplementedException, ConnectionException,
    UnexpectedException)
{
  
  wstring fctName(L"Connection::statementExecuteQuery");

  LockScope ls(&connectionCS);

  checkIfConnected();
  beginTransactionIfNeeded();
  setConnectionParametersOnRequest(request);

  DriverResultSet* retVal = NULL;
  SQLWarning* sqlwPtr = NULL;
  FO_TRY_NTIMES(RECONNECT_RETRIES)
    if (isDebugEnabled())
      logDebug(fctName, L"Sending read request " + static_cast<wstring>(request));
    sendCommand(*driverSocketPtr, StatementExecuteQuery);
    request.sendToStream(*driverSocketPtr);
    // clean up (in case of failover)
    if (retVal != NULL)
      delete retVal;
    if (sqlwPtr != NULL)
      delete sqlwPtr;
    SQLWarning* sqlwPtr = receiveSQLWarnings();
    retVal = receiveResultSet();
    if (retVal != NULL)
      retVal->setStatementWarnings(sqlwPtr);
  FO_CATCH_NTIMES
  return retVal;
}

ResultAndWarnings Connection::statementExecuteUpdate(Request& request)
    throw (SocketIOException, BackendException, ControllerException,
    ProtocolException, ConnectionException, UnexpectedException)
{
  wstring fctName(L"Connection::statementExecuteUpdate");

  LockScope ls(&connectionCS);

  checkIfConnected();
  beginTransactionIfNeeded();
  setConnectionParametersOnRequest(request);

  SQLWarning* sqlwPtr = NULL;
  int controllerResponse = -1;

  // this for-loop is in the case we were disconnected and the controller did
  // not get the update request
  for (int failuresCnt=0; failuresCnt<RECONNECT_RETRIES; failuresCnt++)
  {
    ResultAndWarnings ret;
    FO_TRY_NTIMES(RECONNECT_RETRIES)
      if (isDebugEnabled())
        logDebug(fctName, L"Sending write request " + static_cast<wstring>(request));
      sendCommand(*driverSocketPtr, StatementExecuteUpdate);
      writeRequestOnStream(request);
      request.setId(receiveLongOrException());
    FO_CATCH_NTIMES
  
    // At this point, the request id has been received, let's retrieve the
    // update count
    try
    {
      sqlwPtr = receiveSQLWarnings();    
      controllerResponse = static_cast<int>(receiveIntOrException());
      break; // get out of the big for-loop
    }
    catch (SocketIOException sioe)
    {
      FO_TRY_NTIMES(RECONNECT_RETRIES)
        ret = retrieveExecuteUpdateResult(request);
      FO_CATCH_NTIMES
      if (ret.getUpdateCount() != -1)
      {
        return ret;
      }
      // otherwise, the update was not done on the controller, we go back at
      // the beginning of the function, cleaning up results
      delete ret.getWarnings();
    }
  }

  return ResultAndWarnings(controllerResponse, sqlwPtr);
}

ResultAndWarnings Connection::statementExecute(
    RequestWithResultSetParameters& request) throw (SocketIOException,
    BackendException, ControllerException, ProtocolException,
    ConnectionException, UnexpectedException)
{
  wstring fctName(L"Connection::statementExecute");

  LockScope ls(&connectionCS);
  checkIfConnected();
  beginTransactionIfNeeded();
  setConnectionParametersOnRequest(request);

  SQLWarning* sqlwPtr = NULL;
  std::list<ResultSetOrUpdateCount> results;

  FO_TRY_NTIMES(RECONNECT_RETRIES)
    if (isDebugEnabled())
      logDebug(fctName, L"Sending request " + static_cast<wstring>(request));
    if (!autoCommit)
      writeExecutedInTransaction = true;
    sendCommand(*driverSocketPtr, StatementExecute);
    request.sendToStream(*driverSocketPtr);
    request.setId(receiveLongOrException());
  FO_CATCH_NTIMES

  try
  {
    sqlwPtr = receiveSQLWarnings();
    results = fetchMultipleResultsFromStream();
  }
  catch (SocketIOException sioe)
  {
    FO_TRY_NTIMES(RECONNECT_RETRIES)
      // Clean-up old results if any
      if (results.size() > 0)
      {
        for (std::list<ResultSetOrUpdateCount>::iterator iter = results.begin();
            iter != results.end(); iter++)
        {
          if (iter->isResultSet)
          {
            delete iter->resultSetPtr;
          }
        }
        results.clear();
      }
      if (sqlwPtr != NULL)
      {
        delete sqlwPtr;
        sqlwPtr = NULL;
      }
      // Then retrieve results on controller
      return (retrieveExecuteResult(request));
    FO_CATCH_NTIMES
  }

  return ResultAndWarnings(results, sqlwPtr);
}

ResultAndWarnings Connection::statementExecuteUpdateWithKeys(
    Request &request) throw (SocketIOException, BackendException,
    ControllerException, ProtocolException, ConnectionException,
    UnexpectedException)
{
  wstring fctName(L"Connection::statementExecuteUpdateWithKeys");

  LockScope ls(&connectionCS);

  checkIfConnected();
  beginTransactionIfNeeded();
  setConnectionParametersOnRequest(request);

  SQLWarning* sqlwPtr = NULL;
  std::list<ResultSetOrUpdateCount> result;

  // this for-loop is in the case we were disconnected and the controller did
  // not get the update request
  for (int failuresCnt=0; failuresCnt<RECONNECT_RETRIES; failuresCnt++)
  {
    ResultAndWarnings ret;
    FO_TRY_NTIMES(RECONNECT_RETRIES)
      if (isDebugEnabled())
        logDebug(fctName, L"Executing write request with keys: "
            + static_cast<wstring>(request));
      sendCommand(*driverSocketPtr, StatementExecuteUpdateWithKeys);
      request.sendToStream(*driverSocketPtr);
      request.setId(receiveLongOrException());
    FO_CATCH_NTIMES
    try
    {
      sqlwPtr = receiveSQLWarnings();
      // Receive first the update count or an exception
      ResultSetOrUpdateCount uc;
      uc.isResultSet = false;
      uc.updateCount = static_cast<int>(receiveIntOrException());
      result.push_back(uc);
      // Then the ResultSet containing the autogenerated keys
      ResultSetOrUpdateCount rs;
      rs.isResultSet = true;
      rs.resultSetPtr = receiveResultSet();
      result.push_back(rs);
      break; //no exception, get out of the big for-loop
    }
    catch (SocketIOException sioe)
    {
      FO_TRY_NTIMES(RECONNECT_RETRIES)
        // Clean-up old result if any
        if (result.size() > 0)
        {
          for (std::list<ResultSetOrUpdateCount>::iterator iter = result.begin();
              iter != result.end(); iter++)
          {
            if (iter->isResultSet)
            {
              delete iter->resultSetPtr;
            }
          }
          result.clear();
        }
        if (sqlwPtr != NULL)
        {
          delete sqlwPtr;
          sqlwPtr = NULL;
        }
        // Then retrieve result on controller
        ret = retrieveExecuteUpdateWithKeysResult(request);
      FO_CATCH_NTIMES
      if (ret.getResultList().size() > 1)
        return ret;
      // otherwise, the update was not done on the controller, we go back at
      // the beginning of the function, cleaning up results
      ret.getResultList().clear();
      delete ret.getWarnings();
    }
  }
  return ResultAndWarnings(result, sqlwPtr);
}

void Connection::writeRequestOnStream(const Request& request)
    throw (SocketIOException, UnexpectedException )
{
  if (!autoCommit)
    writeExecutedInTransaction = true;

  request.sendToStream(*driverSocketPtr);

}
void Connection::tryFetchNext(const wstring& cursorName, int fetchSize)
    throw (SocketIOException, ControllerException, ProtocolException,
    BackendException, UnexpectedException)
{   
  wstring fctName(L"Connection::tryFetchNext");

  LockScope ls(&connectionCS);

  checkIfConnected();
  sendCommand(*driverSocketPtr, FetchNextResultSetRows);
  *driverSocketPtr<<cursorName;
  *driverSocketPtr<<fetchSize;
  if (isDebugEnabled())
    logDebug(fctName, L"Fetching next " + toUserString(fetchSize) + L" rows.");
  TypeTag tag(*driverSocketPtr);

  if (tag == TT_EXCEPTION)
    receiveException();

  if (tag != TT_NOT_EXCEPTION)
    throw ProtocolException(L"Protocol exception while trying fetchNext");

  // success, now we can let the DriverResultSet caller receive its data.
}


void Connection::closeRemoteResultSet(wstring cursorName)
    throw (SocketIOException, BackendException, ControllerException,
    ProtocolException, UnexpectedException)
{
  wstring fctName(L"Connection::closeRemoteResultSet");
  
  LockScope ls(&connectionCS);

  checkIfConnected();
  sendCommand(*driverSocketPtr, CloseRemoteResultSet);
  *driverSocketPtr<<cursorName;
  if (isDebugEnabled())
    logDebug(fctName, L"Closing remote ResultSet");

  receiveBoolOrException();
}

bool Connection::receiveBoolOrException() throw (SocketIOException,
                                                 BackendException,
                                                 ControllerException,
                                                 ProtocolException,
                                                 UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);
  //if it's not an exception, receive the bool normally
  if (tag == TT_NOT_EXCEPTION)
  {
    bool answer;
    *driverSocketPtr>>answer;
    return answer;
  }
  //if it's an exception, deserialize and throw it (at the same time)
  receiveException();
  //just to avoid compiler warnings
  return false;
}

int32_t Connection::receiveIntOrException() throw (SocketIOException,
                                                   BackendException,
                                                   ControllerException,
                                                   ProtocolException,
                                                   UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);
  //if it's not an exception, receive the bool normally
  if (tag == TT_NOT_EXCEPTION)
  {
    int32_t answer;
    *driverSocketPtr>>answer;
    return answer;
  }
  //if it's an exception, deserialize and throw it (at the same time)
  receiveException();
  //just to avoid compiler warnings
  return 0;
}

int64_t Connection::receiveLongOrException() throw (SocketIOException,
                                                    BackendException,
                                                    ControllerException,
                                                    ProtocolException,
                                                    UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);
  //if it's not an exception, receive the bool normally
  if (tag == TT_NOT_EXCEPTION)
  {
    int64_t answer;
    *driverSocketPtr>>answer;
    return answer;
  }
  //if it's an exception, deserialize and throw it (at the same time)
  receiveException();
  //just to avoid compiler warnings
  return 0;
}

wstring Connection::receiveStringOrException() throw (SocketIOException,
                                                      BackendException,
                                                      ControllerException,
                                                      ProtocolException,
                                                      UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);
  //if it's not an exception, receive the bool normally
  if (tag == TT_NOT_EXCEPTION)
  {
    wstring answer;
    *driverSocketPtr>>answer;
    return answer;
  }
  //if we are here, it's an exception, deserialize and throw it (at the same
  //time)
  receiveException();
  //just to avoid compiler warnings
  return L"";
}

SQLWarning* Connection::receiveSQLWarnings() throw (SocketIOException,
    BackendException, ControllerException, ProtocolException,
    UnexpectedException)
{
  if (!receiveBoolOrException())
    // no warning
    return NULL;
  TypeTag exceptionType(*driverSocketPtr);
  if (exceptionType != TT_BACKEND_EXCEPTION)
    throw ProtocolException(
        L"Unknown exception received instead of SQLWarning");
  // Receive the warning as a BackendDriverException
  BackendException* e = new BackendException(*driverSocketPtr);
  return convertToSQLWarnings(e);
}

SQLWarning* Connection::getWarnings() throw (DriverException,
    UnexpectedException)
{
  wstring fctName(L"Connection::getWarnings");

  throwExceptionIfClosed(L"Cannot getWarnings() on a closed connection");
  if (!persistent_connection || !retrieve_sql_warnings)
    return first_warning;
  // else...
  FO_TRY_NTIMES(RECONNECT_RETRIES)
    if (isDebugEnabled())
      logDebug(fctName, L"Getting persistent connection warnings");
    sendCommand(*driverSocketPtr, ConnectionGetWarnings);
    *driverSocketPtr << persistent_connection_id;
    return receiveSQLWarnings();
  FO_CATCH_NTIMES
  return NULL; //avoid compiler warning
}

void Connection::clearWarnings() throw (DriverException,
    UnexpectedException)
{
  wstring fctName(L"Connection::clearWarnings");

  if (!persistent_connection || !retrieve_sql_warnings)
    return;
  // else...
  FO_TRY_NTIMES(RECONNECT_RETRIES)
    if (isDebugEnabled())
      logDebug(fctName, L"Getting persistent connection warnings");
    sendCommand(*driverSocketPtr, ConnectionClearWarnings);
    *driverSocketPtr << persistent_connection_id;
    receiveBoolOrException();
  FO_CATCH_NTIMES
}

SQLWarning* Connection::convertToSQLWarnings(BackendException* bde)
{
  if (bde == NULL)
    return NULL;
  // recursion here:
  SQLWarning* sqlw = new SQLWarning(bde->description(), bde->getSQLState(),
      convertToSQLWarnings((BackendException*) bde->getNext()),
      bde->getErrorCode());
  return sqlw;
}

void Connection::throwExceptionIfClosed(const wstring& msg)
    throw (DriverException, UnexpectedException)
{
  if (isClosed)
    throw DriverException(msg);
}

void Connection::receiveException() throw (SocketIOException,
                                           BackendException,
                                           ControllerException,
                                           ProtocolException,
                                           UnexpectedException)
{
  TypeTag exceptionType(*driverSocketPtr);
  if (exceptionType == TT_BACKEND_EXCEPTION)
    throw BackendException(*driverSocketPtr);
  else if (exceptionType == TT_CORE_EXCEPTION)
    throw ControllerException(*driverSocketPtr);
  else throw ProtocolException(L"Received unknown exception type");
}

DriverResultSet* Connection::receiveResultSet()
    throw (SocketIOException, BackendException, ControllerException,
    ProtocolException, NotImplementedException, UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);

  if (tag == TT_NULL_RESULTSET)
    return NULL;

  if (tag == TT_RESULTSET)
  {
    DriverResultSet* drs = new DriverResultSet(this);
    return drs;
  }

  if (tag == TT_EXCEPTION)
    receiveException();

  throw ProtocolException(L"Expected a resultset, received unexpected tag: "
      + toUserString(static_cast<int>(tag)));
  //just to avoid compiler warnings
  return NULL;
}

int64_t Connection::retrieveCommitResult() throw (SocketIOException,
    BackendException, ControllerException, ProtocolException,
    UnexpectedException)
{
  sendCommand(*driverSocketPtr, RetrieveCommitResult);
  return receiveLongOrException();
}

int64_t Connection::retrieveRollbackResult() throw (SocketIOException,
    BackendException, ControllerException, ProtocolException,
    UnexpectedException)
{
  sendCommand(*driverSocketPtr, RetrieveRollbackResult);
  return receiveLongOrException();
}

ResultAndWarnings Connection::retrieveExecuteUpdateResult(const Request &request)
    throw (SocketIOException, BackendException, ControllerException,
    ProtocolException, UnexpectedException)
{
  sendCommand(*driverSocketPtr, RetrieveExecuteUpdateResult);
  *driverSocketPtr << request.getId();
  SQLWarning* sqlwPtr = receiveSQLWarnings();
  int result = static_cast<int>(receiveIntOrException());
  return ResultAndWarnings(result, sqlwPtr);
}

ResultAndWarnings Connection::retrieveExecuteResult(
    const Request &request) throw (SocketIOException, BackendException,
    ControllerException, ProtocolException, UnexpectedException)
{
  sendCommand(*driverSocketPtr, RetrieveExecuteResult);
  *driverSocketPtr << request.getId();
  SQLWarning* sqlwPtr = receiveSQLWarnings();
  std::list<ResultSetOrUpdateCount> result = fetchMultipleResultsFromStream();
  return ResultAndWarnings(result, sqlwPtr);
}

ResultAndWarnings Connection::retrieveExecuteUpdateWithKeysResult(
                        const Request &request) throw (SocketIOException,
                        BackendException, ControllerException,
                        ProtocolException, UnexpectedException)
{
  sendCommand(*driverSocketPtr, RetrieveExecuteUpdateWithKeysResult);
  *driverSocketPtr << request.getId();

  SQLWarning* sqlwPtr = receiveSQLWarnings();
  std::list<ResultSetOrUpdateCount> result;
  // Receive first the update count or an exception
  ResultSetOrUpdateCount uc;
  uc.isResultSet = false;
  uc.updateCount = static_cast<int>(receiveIntOrException());
  if (uc.updateCount != -1)
  {
    result.push_back(uc);
    // Then the ResultSet containing the autogenerated keys
    ResultSetOrUpdateCount rs;
    rs.isResultSet = true;
    rs.resultSetPtr = receiveResultSet();
    result.push_back(rs);
  }
  //else return an empty list
  return ResultAndWarnings(result, sqlwPtr);
}

std::list<ResultSetOrUpdateCount> Connection::fetchMultipleResultsFromStream()
                        throw (SocketIOException, BackendException,
                        ControllerException, ProtocolException,
                        UnexpectedException)
{
  std::list<ResultSetOrUpdateCount> results;
  bool hasResult = false;
  ResultSetOrUpdateCount resultReceived;
  do
  {
    hasResult = receiveBoolOrException();
    if (hasResult)
    {
      resultReceived.isResultSet = true;
      resultReceived.resultSetPtr = receiveResultSet();
//JavaReflecting FIXME: remove this
      results.push_back(resultReceived);
    }
    else
    {
      resultReceived.isResultSet = false;
      resultReceived.updateCount = receiveIntOrException();
//JavaReflecting FIXME: remove this
      if (resultReceived.updateCount != -1)
        results.push_back(resultReceived);
    }
//JavaReflecting FIXME: add this
//    results.push_back(resultReceived);
    
  }
  while (hasResult || resultReceived.updateCount != -1);
  return results;
}

void Connection::reconnectOrDie() throw (ConnectionException,
    DriverException, UnexpectedException)
{
  wstring fctName(L"Connection::reconnectOrDie");
  bool reconnected = false;
  while (!reconnected)
  {
    try
    {
      tryToReconnect();
      reconnected = true;
    }
    catch (ConnectionException ce)
    {
      if (isDebugEnabled())
        logDebug(fctName, L"Controller "
            + static_cast<wstring>(connected_controller)
            + L" seams to be down - giving it a last chance...");
    }
    catch (NoMoreControllerException nmce)
    {
      throw (ConnectionException(nmce.description()));
    }
    // other exceptions are forwarded to caller...
  }
}

void Connection::tryToReconnect() throw (NoMoreControllerException,
DriverException, UnexpectedException)
{
  wstring fctName(L"Connection::tryToReconnect");
  //Wait until other methods/Commands are done
  LockScope ls(&connectionCS);

  if (isWarnEnabled())
    logWarn(fctName, L"Connection to controller lost - trying to reconnect");

  // close everything
  if (isDebugEnabled())
    logDebug(fctName, L"Closing socket of lost connection");

  isClosed = true;
  closeSocket();

  // Add controller to the suspect list...
  if (isDebugEnabled())
    logDebug(fctName, L"Controller " + static_cast<wstring>(connected_controller)
        + L" is not responding, suspecting it of failure");
  ControllerPinger::getInstance().suspectControllerOfFailure(connected_controller);
  
  // ...And try next one in the list
  try
  {
    if (isDebugEnabled())
      logDebug(fctName, L"Getting a new controller from the list...");
    connected_controller = controller_pool.getController();
    initConnection();
    finalizeConnect();
    if (isInfoEnabled())
      logInfo(fctName, L"Connection succeded");
    isClosed = false;
  }
  catch (AuthenticationException ae)
  {
    // Should not happen, this probably mean an inconsistency in controller
    // configuration but safely ignore (see below)
    wstring msg = L"Warning! Authentication exception received on connection retry, controller configuration might be inconsistent";
    if (isWarnEnabled())
      logWarn(fctName, msg);
    throw DriverException(msg+ae.description());
  }

  // Connection succeeded, restore the previous connection state on the
  // controller
  try
  {
    if (isInfoEnabled())
      logInfo(fctName, L"Restoring connection state on the new controller "
          + static_cast<wstring>(connected_controller));
    sendCommand(*driverSocketPtr, RestoreConnectionState);
    *driverSocketPtr<<writeExecutedInTransaction;
    *driverSocketPtr<<autoCommit;
    if (!autoCommit)
      *driverSocketPtr<<transactionId;
    *driverSocketPtr<<persistent_connection;
    if (persistent_connection)
      *driverSocketPtr<<persistent_connection_id;
    *driverSocketPtr<<retrieve_sql_warnings;
    // read ack
    bool ack = false;
    *driverSocketPtr>>ack;
    // Restore read-only state (not part of RestoreConnectionState command for
    // backward compatibility)
    setReadOnly(readOnly);
    if (isInfoEnabled())
      logInfo(fctName, L"Transparent reconnection succeeded");
  }
  catch (SocketIOException sioe)
  {
    throw (ConnectionException(L"Error in reconnection attempt: "
        + sioe.description()));
  }
}
