/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005 Emic Networks
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "ConnectionParameters.hpp"
#include "ControllerPool.hpp"

#include "Common.hpp"

#include <sstream>

using std::wstring;

using namespace CarobNS;

////////////////////////////////////////////////////////////////////////////////
//                             CONTROLLER INFO
////////////////////////////////////////////////////////////////////////////////
ControllerInfo::ControllerInfo() :
    host_name(L""),
    host_port(0)
{
}

ControllerInfo::ControllerInfo(const ControllerInfo& ci) :
    host_name(ci.host_name),
    host_port(ci.host_port)
{
}

ControllerInfo::ControllerInfo(const wstring &name, in_port_t port) :
    host_name(name),
    host_port(port)
{
}

ControllerInfo& ControllerInfo::operator =(const ControllerInfo& ci)
{
  host_name = ci.host_name;
  host_port = ci.host_port;
  return *this;
}

bool ControllerInfo::operator ==(const ControllerInfo& ci) const
{
  return ((host_name == ci.host_name) && (host_port == ci.host_port));
}

bool ControllerInfo::operator <(const ControllerInfo& ci) const
{
  if (host_port < ci.host_port)
    return true;
  if (host_port > ci.host_port)
    return false;
  // compare names if same ports
  return (host_name.compare(ci.host_name) < 0); 
}

bool ControllerInfo::operator >(const ControllerInfo& ci) const
{
  if (host_port > ci.host_port)
    return true;
  if (host_port < ci.host_port)
    return false;
  // compare names if same ports
  return (host_name.compare(ci.host_name) > 0); 
}

ControllerInfo::operator std::wstring() const
{
  std::wostringstream buffer;
  buffer<<host_name<<L":"<<host_port;
  return host_name+L":"+toUserString(static_cast<int>(host_port));
}

////////////////////////////////////////////////////////////////////////////////
//                           CONNECTION PARAMETERS
////////////////////////////////////////////////////////////////////////////////

/**
 * Constants for connection defaults
 */
const std::wstring    ConnectionParameters::DEFAULT_USER                   = L"user"; 
const std::wstring    ConnectionParameters::DEFAULT_PASSWD                 = L"" ;
const std::wstring    ConnectionParameters::DEFAULT_DB                     = L"myDB" ;
const ConnectPolicy   ConnectionParameters::DEFAULT_POLICY                 = ROUND_ROBIN; 
const bool            ConnectionParameters::DEFAULT_CONNECTION_PERSISTENCY = false ;
const bool            ConnectionParameters::DEFAULT_RETRIEVE_SQL_WARNINGS  = false ;

ConnectionParameters::ConnectionParameters(
    const std::vector<ControllerInfo>& ctrls,
    const std::wstring& db /*= DEFAULT_DB*/,
    const std::wstring& uname /*= DEFAULT_USER*/,
    const std::wstring& upass /*= DEFAULT_PASSWD*/,
    ConnectPolicy cp /*= DEFAULT_POLICY*/,
    bool persistentConnections /*= DEFAULT_CONNECTION_PERSISTENCY */,
    bool retrieveSQLWarnings /*= DEFAULT_RETRIEVE_SQL_WARNINGS */)
        throw (ConnectionException, UnexpectedException) :
            database_name(checkDatabaseName(db)),
            user_name(uname),
            user_pass(upass),
            connect_policy(cp),
            persistent_connections(persistentConnections),
            retrieve_sql_warnings(retrieveSQLWarnings)
{
  // Check the given controllers host names
  for (std::vector<ControllerInfo>::const_iterator iter = ctrls.begin();
      iter != ctrls.end(); iter++)
  {
    checkHostName(iter->getHostName());
  }
  controller_list = ctrls;
}

ConnectionParameters::~ConnectionParameters()
{
}

AbstractControllerPool& ConnectionParameters::getControllerPool()
{
  controller_pool = &ControllerPoolManager::getPool(controller_list, connect_policy);
  return *controller_pool;
}

void ConnectionParameters::releaseControllerPool()
{
  ControllerPoolManager::freePool(controller_pool);
}

void ConnectionParameters::checkHostName(const wstring& hostNamePrm) const
    throw (ConnectionException, UnexpectedException)
{
  wchar_t pbChar;
  if (!isValidHostName(hostNamePrm, pbChar))
  {
    wstring msg = L"Invalid hostname " + hostNamePrm + L" at character '"
        + static_cast<wchar_t>(pbChar) + L"'.";
    throw ConnectionException(msg);
  }
}

const wstring& ConnectionParameters::checkDatabaseName(const wstring& dbname) const
    throw (ConnectionException, UnexpectedException)
{
  wchar_t pbChar;
  if (!isValidDatabaseName(dbname, pbChar))
      throw ConnectionException(L"Invalid database name "+dbname+L" at character '"+pbChar+L"'.");
  return dbname;
}

bool ConnectionParameters::isValidDatabaseName(const wstring& databaseName, wchar_t& pbChar) const
{
  int size = databaseName.length();
  wchar_t c;

  for (int i = 0; i < size; i++)
  {
    c = databaseName[i];
    if ((c < L'-') || (c > L'z') || (c == L'/') || (c == L'.') || (c == L'`')
        || ((c > L'9') && (c < L'A')) || ((c > L'Z') && (c < '_')))
    {
      pbChar = c;
      return false;
    }
  }
  return true;
}

bool ConnectionParameters::isValidHostName(const wstring& hostNamePrm,
    wchar_t& pbChar) const
{
  int size = hostNamePrm.length();
  wchar_t c;
  wchar_t lastChar = L' ';

  for (int i = 0; i < size; i++)
  {
    c = hostNamePrm[i];

    if (c == L'.' || c == L'-')
    {
      if (lastChar == L'.' || lastChar == L'-' || (i == size - 1) || (i == 0))
      {
        // . or - cannot be the first or the last char of hostname
        // hostname cannot contain '..' or '.-' or '-.' or '--'
        pbChar = c;
        return false;
      }
    }
    else
    {
      if (((c < L'0') || (c > L'z') || ((c > L'9') && (c < L'A'))
          || ((c > L'Z') && (c < L'_')) || (c == L'`')))
      {
        pbChar = c;
        return false;
      }
    }
    lastChar = c;
  }
  return true;
}


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
