/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "ControllerPool.hpp"

#include "ControllerPinger.hpp"
#include "JavaSocket.hpp"
#include "Connection.hpp"

#include "ConnectionParameters.hpp"
#include "Common.hpp"

#include <map>
#include <sstream>

using std::wstring;

using namespace CarobNS;


/* Pool index less than operator */
bool ltPoolIndex::operator()(const PoolIndex& pi1, const PoolIndex& pi2) const
{
  if (pi1.policy < pi2.policy)
    return true;
  if (pi1.policy > pi2.policy)
    return false;
  // same policy, compare number of controllers
  if (pi1.controllers.size() < pi2.controllers.size())
    return true;
  else if (pi1.controllers.size() < pi2.controllers.size())
    return false;
  // same size, compare controllers themselves
  std::vector<ControllerInfo>::const_iterator iter_pi1 = pi1.controllers.begin();
  std::vector<ControllerInfo>::const_iterator iter_pi2 = pi2.controllers.begin();
  while (iter_pi1 != pi1.controllers.end())
  // we can safely check only pi1, vectors have same size
  {
    if (*iter_pi1 < *iter_pi2)
      return true;
    if (*iter_pi1 > *iter_pi2)
      return false;
    iter_pi1++;
    iter_pi2++;
  } 
  // identical
  return false;
}

////////////////////////////////////////////////////////////////////////////////
//                           ABSTRACT CONTROLLER POOL
////////////////////////////////////////////////////////////////////////////////
AbstractControllerPool::AbstractControllerPool(
    const std::vector<ControllerInfo>& controllerList, long retryIntervalInMs)
        throw (DriverException, UnexpectedException) : ref_counter(0)
{
  if (controllerList.size() == 0)
    throw DriverException(
        L"Controller list must at least contain one controller!");
  controller_list = controllerList;
}

AbstractControllerPool::~AbstractControllerPool()
{
  controller_list.clear();
}


AbstractControllerPool::operator wstring() const
{
  wstring sRet(L"Pool controller list: ");
  for (size_t i = 0; i < controller_list.size(); i++)
  {
    sRet += L"[" + static_cast<wstring>(controller_list[i]) + L"]";
  }
  sRet += L" - Ref counter=" + toUserString(ref_counter);
  return sRet;
}

////////////////////////////////////////////////////////////////////////////////
//                             ROUND ROBIN POLICY
////////////////////////////////////////////////////////////////////////////////

RoundRobinControllerPool::RoundRobinControllerPool(
    const std::vector<ControllerInfo>& controllerList, long retryIntervalInMs)
        throw (DriverException, UnexpectedException) :
AbstractControllerPool(controllerList, retryIntervalInMs),
index(-1)
{
}

ControllerInfo RoundRobinControllerPool::getController()
    throw (NoMoreControllerException, UnexpectedException)
{
  wstring fctName(L"RoundRobinControllerPool::getController");

  LockScope ls(&pool_CS);
  {
    unsigned int testedControllers = 0; //unsigned because we compare it to vector.size()
    bool lastTestedControllerWasSuspect = false;
    do
    {
      index = (index + 1) % controller_list.size();
      lastTestedControllerWasSuspect = 
          ControllerPinger::getInstance().isSuspectedOfFailure(controller_list[index]);
      testedControllers++;
    }
    while (lastTestedControllerWasSuspect
        && testedControllers <= controller_list.size());
    // if there are no more controllers up => exception
    if (lastTestedControllerWasSuspect)
    {
      throw NoMoreControllerException(L"All "
                                      + toUserString(controller_list.size())
                                      + L" controllers down");
    }
  }
  if (isDebugEnabled())
    logDebug(fctName, L"Selected controller[" + toUserString(index) + L"]:"
        + static_cast<wstring>(controller_list[index]));
  return controller_list[index];
}

////////////////////////////////////////////////////////////////////////////////
//                           CONTROLLER POOL MANAGER
////////////////////////////////////////////////////////////////////////////////
/* static members declaration */
CriticalSection ControllerPoolManager::pool_map_CS;
std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>  ControllerPoolManager::pool_map;

/*static*/
AbstractControllerPool& ControllerPoolManager::getPool(std::vector<ControllerInfo> ctrls,
    ConnectPolicy cp) throw (DriverException, UnexpectedException)
{
  // iterate through the list to find the wanted pool, if present
  PoolIndex idx;
  idx.controllers = ctrls;
  idx.policy = cp;
  LockScope ls(&pool_map_CS);
  std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>::iterator pool
      = pool_map.find(idx);
  if (pool != pool_map.end())
  {
    pool->second->addRef();
    return *(pool->second);
  }
  // if the pool was not found, let's create a new one
  AbstractControllerPool* new_pool;
  switch (cp)
  {
    case ROUND_ROBIN:
      new_pool = new RoundRobinControllerPool(ctrls);
    break;
    default:
      throw DriverException(L"Unsupported connection policy #" + toUserString(static_cast<int>(cp)));
  }
  new_pool->addRef();
  pool_map[idx] = static_cast<AbstractControllerPool*>(new_pool);
  return *new_pool;
}

/*static*/
void ControllerPoolManager::freePool(AbstractControllerPool* pool)
{
  wstring fctName(L"ControllerPoolManager::freePool");
  if (pool == NULL)
  {
    if (isErrorEnabled())
      logError(fctName, L"Null pool can't be freed!");
    return;
  }
  for (std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>::iterator iter
      = pool_map.begin(); iter != pool_map.end(); iter++)
  {
    if (iter->second == pool)
    {
      if (pool->removeRef() == 0)
      {
        //iter->second = NULL;
        // delete pool will be called by erase
        pool_map.erase(iter); // don't mind about iter, we won't use it anymore
        delete pool;
      }
      return;
    }
  }
  // No match => log error and continue
  if (isErrorEnabled())
    logError(fctName, L"Pool " + static_cast<wstring>(*pool)
        + L" not found in pool map!");
}
