/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Emic Networks
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon
 */

#include "JavaSocket.hpp"

#include "Common.hpp"
#include "SystemDependantDefs.hpp"

#ifdef __MINGW32__
  #include <ws2tcpip.h>
  #define __BYTE_ORDER BYTE_ORDER
#else
  #include <netdb.h> // getaddrinfo() etc.
  #include <netinet/tcp.h> // TCP_NODELAY etc.
  #include <unistd.h> // sleep()
#endif
#include <errno.h>
#include <fcntl.h>

using std::wstring;

using namespace CarobNS;

JavaSocket::JavaSocket() : 
socketFd(-1),
connected(false)
{
}

JavaSocket::~JavaSocket() throw (SocketIOException, UnexpectedException)
{
  wstring fctName(L"JavaSocket::~JavaSocket");
  if (isDebugEnabled())
    logDebug(fctName, L"About to close socket...");
  closeSocket();
}

bool JavaSocket::create(bool blocking /* = true */)
    throw (ConnectionException, UnexpectedException)
{
  wstring fctName(L"JavaSocket::Create");
  if (isDebugEnabled())
    logDebug(fctName, L"Creating socket...");
  socketFd = socket(SOCKET_CREATION_DOMAIN,
                    SOCK_STREAM,
                    0);

  if (socketFd == -1)
  {
    if (isErrorEnabled())
      logError(fctName, L"Socket creation failed");
    throw ConnectionException(L"Socket creation failed");
    return false;
  }

  if (!blocking)
  {
#ifdef __MINGW32__
    int var = 1;
    ioctlsocket(socketFd, FIONBIO, (u_long FAR*) &var);
    if (var == 0)
    {
      if (isErrorEnabled())
        logError(fctName, L"Could not turn socket to non-blocking mode.");
    }
#else
    // *nix:
    int flags;
    flags = fcntl(socketFd, F_GETFL, 0);
    //Set non blocking
    if (fcntl(socketFd, F_SETFL, flags | O_NONBLOCK) < 0)
    //if above fcntl doesn't work, try this old way of doing it
    //flags = 1;
    //if (ioctl(socketFd, FIOBIO, &flags) < 0)
    {
      if (isErrorEnabled())
        logError(fctName, L"Could not turn socket to non-blocking mode");
    }
#endif
  }
  //Set socket options
  int opt_value = 1;
  if (setsockopt(socketFd, IPPROTO_TCP, TCP_NODELAY,
                 reinterpret_cast<char*>(&opt_value), sizeof opt_value) == -1
#if (defined(__MACH__) && defined(__APPLE__)) // quick & dirty fix for MacOSX
      || setsockopt(socketFd, SOL_SOCKET, SO_NOSIGPIPE,
                    reinterpret_cast<char*>(&opt_value), sizeof (opt_value) == -1)
#endif
    )
  {
    if (isErrorEnabled())
      logError(fctName, L"Socket set option failed");
    throw ConnectionException(L"Set option failed on socket");
    return false;
  }

  // Turn keepalive mechanism on 
  int keepalive = 1;
  if (setsockopt(socketFd, SOL_SOCKET, SO_KEEPALIVE, &keepalive, sizeof(keepalive)) < 0)
  {
    if (isErrorEnabled())
      logError(fctName, L"Could not set socket keepalive flag");
  }

  if (isDebugEnabled())
    logDebug(fctName, L"Socket created.");
  return true;
}

bool JavaSocket::connectTo(const std::wstring& host, in_port_t port)
    throw (ConnectionException, UnexpectedException)
{
  wstring fctName(L"JavaSocket::connectTo");
  if (isDebugEnabled())
    logDebug(fctName, L"Connecting...");
  if (!isValid())
  {
    if (isErrorEnabled())
      logError(fctName, L"Invalid socket. Did you Create() it ?");
    throw ConnectionException(L"Invalid socket for connection");
    connected = false;
    return false;
  }

  // getaddrinfo() is IPV6 compatible and will give us a *list* of valid 
  // addresses. This is why we don't give 'hints' parameter (we want
  // all valid addresses).
  // In the 'normal' case, there will be only one address to 
  // connect to, but if we get multiple addresses, we will try to connect 
  // to all of these (in the order returned by getaddrinfo()
  
  struct addrinfo *addressInfoList = NULL,
                  *addressInfoPtr = NULL;
  sockaddr_in addr = {0};
  // try
  std::string hostAsString = toString(host);

  int error = getaddrinfo(hostAsString.c_str(), NULL, NULL, &addressInfoList);
  
  if (error)
  {
#ifdef __MINGW32__
    wstring wMsg = L"Error while executing getaddrinfo().";
#else
    const char* gaiMsg = gai_strerror(error);
    wstring wMsg = fromString(gaiMsg);
#endif
    if (isDebugEnabled())
      logDebug(fctName, wMsg);
    connected = false;
    throw ConnectionException(wMsg);
    return false;
  }
  else
  {
    addressInfoPtr = addressInfoList;
    while (addressInfoPtr != NULL)
    {
      addr = *(reinterpret_cast<struct sockaddr_in*>(addressInfoList->ai_addr));
      addr.sin_port = htons(port);
      
      int resp = ::connect(socketFd, reinterpret_cast<sockaddr*>(&addr), sizeof(addr));
      if (resp==0)
      {
        if (isInfoEnabled())
          logInfo(fctName, L"Connection to "+host+L":"+toUserString(port)+L" succeeded");
        freeaddrinfo(addressInfoList);
        connected = true;
        return true;
      }
      else
      {
        if (isDebugEnabled())
          logDebug(fctName, L"Could not connect trying next address if any...");
      }
      addressInfoPtr = addressInfoPtr->ai_next;
    }
  }
  //if we get here, it means we could'nt connect to any of the addresses found...
  freeaddrinfo(addressInfoList);
  throw ConnectionException(L"Unable to connect. Last error code was "+toUserString(errno));
  connected = false;
  return false;
}

bool JavaSocket::closeSocket() throw (SocketIOException, UnexpectedException)
{
  wstring fctName(L"closeSocket");
  if (isValid())
  {
#ifdef __MINGW32__
    if (closesocket(socketFd) != 0)
#else
    if (close(socketFd) != 0)
#endif
    {
      wstring msg(L"Could not close socket. Error code is ");
#ifdef __MINGW32__
      msg += toUserString(WSAGetLastError());
#else
      msg += toUserString(errno);
#endif
      throw SocketIOException(msg);
      return false;
    }
    if (isDebugEnabled())
      logDebug(fctName, L"Socket closed.");
  }
  socketFd = -1;
  connected = false;
  return true;
}

// Localization (Chapter 22) in
// /usr/share/doc/gcc-4.0-base/libstdc++/html/documentation.html

size_t JavaSocket::writeJavaUTF(const wstring& str) const
  throw (SocketIOException, CodecException, UnexpectedException)
{
  wstring fctName(L"JavaSocket::writeJavaUTF");

  std::string utf8str(toUTF8(str));

  int32_t netlen = htonl(utf8str.length());

  //First write number of bytes to follow as 
  sendToSocket(fctName, L"UTF string", &netlen, sizeof(netlen), SOCKET_SEND_FLAGS);
  sendToSocket(fctName, L"UTF string", utf8str.data(), utf8str.length(), SOCKET_SEND_FLAGS);

  return utf8str.length();
}

size_t JavaSocket::readJavaUTF(wstring& s) const
  throw (SocketIOException, CodecException, UnexpectedException)
{
  wstring fctName(L"JavaSocket::readJavaUTF");
  
  int32_t lenRecNet;
  //the converted size
  uint16_t lenRec;
  size_t sizeRead = 0;
  //First get size
  receiveFromSocket(fctName, L"UTF string size", &lenRecNet, sizeof(lenRecNet), 0);
  lenRec = ntohl(lenRecNet); // number of bytes to come

  uint8_t* utfStr = new uint8_t[lenRec];
  receiveFromSocket(fctName, L"UTF string", utfStr, lenRec, 0);
  const std::string received(reinterpret_cast<char*>(utfStr), lenRec);

  try {
    s = fromUTF8(received);
  } catch (CodecException) {
    delete[] utfStr; throw;
  }
  sizeRead = (size_t)lenRec;

  delete[] utfStr;
  return sizeRead;
}

void JavaSocket::writeJavaInt(int32_t i) const throw (SocketIOException,
    UnexpectedException)
{
  wstring fctName(L"JavaSocket::WriteInt32");
  // we have to convert the byte order
  int32_t conv = htonl(i);
  sendToSocket(fctName, L"Int32", &conv, sizeof(conv), SOCKET_SEND_FLAGS);
}

void JavaSocket::readJavaInt(int32_t& i) const throw (SocketIOException,
    UnexpectedException)
{
  wstring fctName(L"Socket::ReadInt32");
  // the data will be received with network byte order.
  // so we must convert it after reception
  int32_t rec = 0;
  
  receiveFromSocket(fctName, L"Int32", &rec, sizeof(rec), 0);
  i = ntohl(rec);
}

void JavaSocket::writeJavaLong(int64_t i) const throw (SocketIOException,
    UnexpectedException)
{
  wstring fctName(L"JavaSocket::writeJavaLong");
  // we have to convert the byte order
  int64_t conv = htonll(i);
  sendToSocket(fctName, L"Int64", &conv, sizeof(conv), SOCKET_SEND_FLAGS);
}

void JavaSocket::readJavaLong(int64_t& i) const throw (SocketIOException,
    UnexpectedException)
{
  wstring fctName(L"Socket::readJavaLong");
  // the data will be received with network byte order.
  // so we must convert it after reception
  int64_t rec = 0;
  
  receiveFromSocket(fctName, L"Int64", &rec, sizeof(rec), 0);
  i = ntohll(rec);
}

void JavaSocket::writeJavaBool(bool b) const throw (SocketIOException,
    UnexpectedException)
{
  wstring fctName(L"JavaSocket::writeJavaBool");
  uint8_t byteToSend = b; // the implicit casts do it right
  //No byte order problem here: the only value read is '0' for false
  //all other values are true
  sendToSocket(fctName, L"Boolean", &byteToSend, sizeof(byteToSend), SOCKET_SEND_FLAGS);
}

void JavaSocket::readJavaBool(bool& b) const throw (SocketIOException,
    UnexpectedException)
{
  wstring fctName(L"JavaSocket::readJavaBool");

  uint8_t rec;
  receiveFromSocket(fctName, L"Boolean", &rec, sizeof(rec), 0);
  b = rec; // the implicit cast does it right
}

void JavaSocket::readJavaBytes(int32_t length, java_byte* data) const
    throw (SocketIOException, UnexpectedException)
{
  wstring fctName(L"JavaSocket::readJavaBytes");
  receiveFromSocket(fctName, L"JavaBytes", data, length, 0);
}

void JavaSocket::writeJavaBytes(int32_t length, java_byte* data) const
    throw (SocketIOException, UnexpectedException)
{
  wstring fctName(L"JavaSocket::writeJavaBytes");
  sendToSocket(fctName, L"JavaBytes", data, length, SOCKET_SEND_FLAGS);
}

/** Returns the number of octets read or the (negative) error from recv() */
int JavaSocket::recvFully(void *buf, const int len, const int flags)
    const
{
  int alreadyRead = 0; //Keeps the whole size read 
  int readThisTime = 0; //The size read at each receive
  while (alreadyRead < len)
  {
    readThisTime = recv(socketFd,
                        (char*)(static_cast<uint8_t*>(
                                      ((static_cast<uint8_t*>(buf))+alreadyRead))),
                        len-alreadyRead,
                        flags);
    if (readThisTime <= 0)
    {
      //this is an error
      return readThisTime;
    }
    alreadyRead += readThisTime;
  }
  return alreadyRead;
}

void JavaSocket::sendToSocket(const wstring& fctName, const wstring& objName,
    const void* buf, int len, int flags) const
    throw (SocketIOException, UnexpectedException)
{
  // in non blocking mode, send will block until len bytes are send. So no need
  // to loop here
  int ret = send(socketFd, static_cast<const char*>(buf), len, flags);
  if ( ret == -1 )
  {
    throw SocketIOException(L'(' + fctName + L"): could not write "
        + objName + L" to socket");
  }
  if (ret != len)
  {
    // should never happen
    throw SocketIOException(L'(' + fctName + L"): only " + toUserString(ret)
        + L"/" + toUserString(len) + L" bytes send while sending " + objName
        + L" to socket");
  }
}

void JavaSocket::receiveFromSocket(const wstring& fctName,
    const wstring& objName, void *buf, int len, int flags) const
    throw (SocketIOException, UnexpectedException)
{
  if (0 == len)
    return;

  int status = recvFully(buf, len, flags);

  if (status == -1)
    throw SocketIOException(fctName + L"Could not read from socket.");

  if (status == 0)
    throw SocketIOException(fctName + L"Peer reset connection while receiving");

  // else it went fine
  // if (status != len)
  //  throw SocketIOException(fctName + L"something is really wrong!");
}

const uint64_t JavaSocket::ntohll(const uint64_t &n) const
{
#ifndef __BYTE_ORDER // typically found through <sys/param.h> 
                     // or <netinet/*.h>
#error "__BYTE_ORDER undefined: unknown endianness, can't implement 64bits swaps"
#endif
#if __BYTE_ORDER == __BIG_ENDIAN
  return n;
#else
  return (((uint64_t)ntohl(n)) << 32) + ntohl(n >> 32);
#endif
}

const uint64_t JavaSocket::htonll(const uint64_t &n) const
{
#if __BYTE_ORDER == __BIG_ENDIAN
  return n;
#else
  return (((uint64_t)htonl(n)) << 32) + htonl(n >> 32);
#endif
}

/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */
