/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Marc Herbert
 * Contributor(s):
 */

#include "StringCodecs.hpp"

#include "Common.hpp"

#include <stdexcept>
#include <sstream>

using namespace CarobNS;

namespace {

// Warning: this function needs the codec of fromString() to have been
// properly initialized.
std::wstring
exception_message(const char *what, const char *locname)
{
    std::wstring msg(fromString(what));
    msg += fromString(", locale name was: ") + fromString(locname);
    return msg;
}

}

/**
 * Catching std::runtime_error from std::codecvt_byname() segfaults
 * with gcc 4.0.2, see
 * <http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=343108>
 * So unfortunately do NOT do this. This code is a reminder.
 */
namespace {
template <typename facet_byname_t>
facet_byname_t * trynewfacet_byname(const char * const name)
{
    try {
        return new facet_byname_t(name);
    } catch (std::runtime_error& re) {
        // we need the default user_codec to be initialized here
        std::wstring msg = exception_message(re.what(), name);
        logFatal(__WFILE__ L":trynewfacet_byname()", msg);
        throw CodecException(msg);
    }
}
}


std::locale CarobNS::trylocale(const char * const name) throw (CodecException)
{
    try {
        return std::locale(name);
    } catch (std::runtime_error& re) {
        // we need the default user_codec to be initialized here
        std::wstring msg = exception_message(re.what(), name);
        logWarn(__WFILE__ L":trylocale()" , msg);
        throw CodecException(msg);
    }
}


std::string MBSCodec::encode(const std::wstring& w_arg) const throw (CodecException)
{
    mbstate_t mbst = { 0 };

    // prefered to cvt.length(in), assumed to be too costly cause of the scan
    // (it's a memory/CPU tradeoff)
    int res_maxlength = codecvt.max_length() * w_arg.length();
    char* res_buf = new char[res_maxlength];

    const wchar_t *w_argnext;
    char *res_next;
    const wchar_t *w_argdata = w_arg.data();

    std::codecvt_base::result success =
        codecvt.out(mbst, w_argdata, w_argdata+w_arg.length(), w_argnext,
                res_buf, res_buf+res_maxlength, res_next);

    if (success != std::codecvt_base::ok) // codecvt_base::noconv is not possible
    {
        std::wostringstream woss;
        woss << L"Error " << success << L" in "
                  << fromString(loc.name()) << L" encoding method";
        logError(__WFILE__ L"MBSCodec::encode", woss.str());
        delete[] res_buf;
        throw CodecException(woss.str());
    }
    std::string retStr(res_buf, res_next-res_buf);
    delete[] res_buf;
    return retStr;
}


std::wstring MBSCodec::decode(const std::string& arg) const throw (CodecException)
{
    mbstate_t mbst = { 0 };

    int arglength = arg.length();
    wchar_t* wres_buf = new wchar_t[arglength];
    
    const char *argnext;
    wchar_t *wres_next;
    const char *argdata = arg.data();

    std::codecvt_base::result success =
        codecvt.in(mbst, argdata, argdata+arglength, argnext,
                   wres_buf, wres_buf+arglength, wres_next);

    if (success != std::codecvt_base::ok)  // codecvt_base::noconv is not possible
    {
        std::wostringstream woss;
        woss << L"Error " << success << L" in "
                  << fromString(loc.name()) << L" decoding method";
        logError(__WFILE__ L"MBSCodec::decode()", woss.str());
        delete[] wres_buf;
        throw CodecException(woss.str());
    }
    std::wstring retWStr(wres_buf, wres_next-wres_buf);
    delete[] wres_buf;
    return retWStr;
}


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
