/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

// not to derive: just to get default parameters
#include "../ConnectionSetup.hpp"

#include "Connection.hpp"
#include "Common.hpp"
#include "ConnectionParameters.hpp"
#include "CarobException.hpp"
#include "TestConnect.hpp"

using std::wstring;

using namespace CarobNS;

void TestConnect::setUp()
{
  connectionPtr = NULL;
}

void TestConnect::tearDown()
{
  if (connectionPtr)
    delete connectionPtr;
}

void TestConnect::testConnectBadAddress()
{
  wstring fctName(L"TestConnect::testConnectBadAddress");
  std::vector<ControllerInfo> ctrls;
  ctrls.push_back(ControllerInfo(L"dummyAddress", ConnectionSetup::DEFAULT_PORT));
  ConnectionParameters connectionPrms(ctrls,
                                      L"myDB",
                                      L"user",
                                      L"");  
  try
  {
    if (isInfoEnabled())
    {
      logInfo(fctName, L"Connecting to controller - should fail");
    }
    connectionPtr = new Connection(connectionPrms);
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (ConnectionException ce)
  {
    if (isErrorEnabled())
      logError(fctName, L"Connection failed (this is ok). Exception: "
          + ce.description());
    CPPUNIT_ASSERT(true);
  }
}

void TestConnect::testConnectBadPort()
{
  wstring fctName(L"TestConnect::testConnectBadPort");
  std::vector<ControllerInfo> ctrls;
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 12345));
  ConnectionParameters connectionPrms(ctrls,
                                      L"myDB",
                                      L"user",
                                      L"");  
  try
  {
    if (isInfoEnabled())
    {
      logInfo(fctName, L"Connecting to controller - sould fail");
    }
    connectionPtr = new Connection(connectionPrms);
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (ConnectionException ce)
  {
    if (isErrorEnabled())
      logError(fctName, L"Connection failed (this is ok). Exception: "
          + ce.description());
    CPPUNIT_ASSERT(true);
  }
}
void TestConnect::testConnectBadDB()
{
  wstring fctName(L"TestConnect::testConnectBadDB");
  std::vector<ControllerInfo> ctrls;
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, ConnectionSetup::DEFAULT_PORT));
  ConnectionParameters connectionPrms(ctrls,
                                      L"dummyDB",
                                      L"user",
                                      L"");  
  try
  {
    connectionPtr = new Connection(connectionPrms);
    if (isInfoEnabled())
    {
      logInfo(fctName, L"Connecting to controller - sould fail");
    }
    
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (AuthenticationException ae)
  {
    if (isErrorEnabled())
      logError(fctName, L"Connection failed (this is ok). Exception: "
          + ae.description());
    CPPUNIT_ASSERT(true);
  }
}

void TestConnect::testConnectBadUser()
{
  wstring fctName(L"TestConnect::testConnectBadUser");
  std::vector<ControllerInfo> ctrls;
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, ConnectionSetup::DEFAULT_PORT));
  ConnectionParameters connectionPrms(ctrls,
                                      L"myDB",
                                      L"dummyuser",
                                      L"");  
  try
  {
    connectionPtr = new Connection(connectionPrms);
    if (isInfoEnabled())
    {
      logInfo(fctName, L"Connecting to controller - sould fail");
    }
    
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (AuthenticationException ae)
  {
    if (isErrorEnabled())
      logError(fctName, L"Connection failed (this is ok). Exception: "
          + ae.description());
    CPPUNIT_ASSERT(true);
  }
}

void TestConnect::testConnectBadPass()
{
  wstring fctName(L"TestConnect::testConnectBadPass");
  std::vector<ControllerInfo> ctrls;
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, ConnectionSetup::DEFAULT_PORT));
  ConnectionParameters connectionPrms(ctrls,
                                      L"myDB",
                                      L"user",
                                      L"dummyPass");  
  try
  {
    connectionPtr = new Connection(connectionPrms);
    if (isInfoEnabled())
    {
      logInfo(fctName, L"Connecting to controller - sould fail");
    }
    
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (AuthenticationException ae)
  {
    if (isErrorEnabled())
      logError(fctName, L"Connection failed (this is ok). Exception: "
          + ae.description());
    CPPUNIT_ASSERT(true);
  }
}

void TestConnect::testConnectGood()
{
  wstring fctName(L"TestConnect::testConnectGood");
  std::vector<ControllerInfo> ctrls;
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, ConnectionSetup::DEFAULT_PORT));
  ConnectionParameters connectionPrms(ctrls,
                                      L"myDB",
                                      L"user",
                                      L"");  
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Connecting to controller - this should succeed");
  }
  connectionPtr = new Connection(connectionPrms);
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Connection succeeded");
  }
}

CppUnit::Test* TestConnect::suite()
{
  CppUnit::TestSuite *suiteOfTests = new CppUnit::TestSuite( "TestConnect" );
  suiteOfTests->addTest(new CppUnit::TestCaller<TestConnect>(
                                 "TestConnect::testConnectBadAddress",
                                 &TestConnect::testConnectBadAddress));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestConnect>(
                                 "TestConnect::testConnectBadPort",
                                 &TestConnect::testConnectBadPort));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestConnect>(
                                 "TestConnect::testConnectBadDB", 
                                 &TestConnect::testConnectBadDB));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestConnect>(
                                 "TestConnect::testConnectBadUser", 
                                 &TestConnect::testConnectBadUser));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestConnect>(
                                 "TestConnect::testConnectBadPass", 
                                 &TestConnect::testConnectBadPass));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestConnect>(
                                 "TestConnect::testConnectGood", 
                                 &TestConnect::testConnectGood));

  return suiteOfTests;
}
