/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "TestFailOver.hpp"

// not to derive: just to get default parameters
#include "../ConnectionSetup.hpp"

#include "Statement.hpp"
#include "Connection.hpp"

#include "Common.hpp"

#include <string>
#include <iostream>

using std::wstring;

#ifdef __MINGW32__
  #define sleep _sleep
#endif

using namespace CarobNS;

void TestFailOver::setUp()
{
}

void TestFailOver::tearDown()
{
}

void TestFailOver::testReconnectNoControllers()
{
  std::vector<ControllerInfo> ctrls;

  // this is a dummy port
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 12345));
  ConnectionParameters connectionPrms(ctrls,
                                      L"myDB",
                                      L"user",
                                      L"");
  // 1st try with only 1 failing ctrler
  try
  {
    Connection* connectionPtr = new Connection(connectionPrms);
    connectionPtr->setAutoCommit(false); //just to avoid unused variable warning
  }
  catch (ConnectionException ce)
  {
    if (isInfoEnabled())
      logError(L"TestFailOver::testReconnectNoControllers",
          L"new Connection throw connection exception (this is ok): "
          + ce.description());
  }
  // 2nd try with 2 (first one is now suspected of failure...
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 12346));
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 12347));
  try
  {
    Connection* connectionPtr = new Connection(connectionPrms);
    connectionPtr->setAutoCommit(false); //just to avoid unused variable warning
  }
  catch (ConnectionException ce)
  {
    if (isInfoEnabled())
      logError(L"TestFailOver::testReconnectNoControllers",
          L"new Connection throw connection exception (this is ok): "
          + ce.description());
  }
}

void TestFailOver::testConnectOnlyOneControllerUp()
{
  std::vector<ControllerInfo> ctrls;

  // this is a dummy port
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 12345));
  // 2nd controller is down too
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 12349));
  // 3rd controller is ok
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 25322));
  // 4th is down
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 12350));
  ConnectionParameters connectionPrms(ctrls,
                                      L"myDB",
                                      L"user",
                                      L"");
  Connection* connectionPtr = new Connection(connectionPrms);
  //no exception.. good. clean the connection
  delete connectionPtr;
}

void TestFailOver::testReconnect()
{
  std::vector<ControllerInfo> ctrls;

  // this is a dummy controller
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 25324));
  // this one will work
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 25322));
  // this one too
  ctrls.push_back(ControllerInfo(ConnectionSetup::DEFAULT_HOST, 25323));
  ConnectionParameters connectionPrms(ctrls,
                                      L"myDB",
                                      L"user",
                                      L"");
  Connection* connectionPtr = new Connection(connectionPrms);
  Statement* statementPtr = connectionPtr->createStatement();

  wstring fctName(L"TestExecWriteRequest::testReconnect");
  wstring requestBegin = L"UPDATE product SET cost=";
  wstring requestEnd = L" where id=";

//  int nbRowsAffected;
  bool rsuc;
  if (isInfoEnabled())
    logInfo(fctName, L"Executing 1000 writes");
  for (int i=0; i<1000; i++)
  {
    try
    {
//      nbRowsAffected = statementPtr->executeUpdate(
//          requestBegin + toWString(i) + requestEnd + toWString(i%50));
      rsuc = statementPtr->execute(
          requestBegin + toWString(i) + requestEnd + toWString(i%50));
    }
    catch (NoMoreControllerException nmce)
    {
      if (isInfoEnabled())
          logError(fctName, L"Write failed (this is ok). Exception:"
              + nmce.description());      
    }
//    CPPUNIT_ASSERT(nbRowsAffected == 1);
    CPPUNIT_ASSERT(rsuc == false);
    sleep(1);
  }
}

CppUnit::Test* TestFailOver::suite()
{
  CppUnit::TestSuite *suiteOfTests = new CppUnit::TestSuite( "TestFailOver" );
  suiteOfTests->addTest(new CppUnit::TestCaller<TestFailOver>(
                                 "TestFailOver::testReconnectNoControllers",
                                 &TestFailOver::testReconnectNoControllers));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestFailOver>(
                                 "TestFailOver::testConnectOnlyOneControllerUp",
                                 &TestFailOver::testConnectOnlyOneControllerUp));
//disabled test: needs user interaction
/*  suiteOfTests->addTest(new CppUnit::TestCaller<TestFailOver>(
                                 "TestFailOver::testReconnect",
                                 &TestFailOver::testReconnect));
*/

  return suiteOfTests;
}
