/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "TestDriverResultSet.hpp"

#include "Statement.hpp"
#include "DriverResultSet.hpp"

#include "CarobException.hpp"
#include "Common.hpp"

#include <string>
#include <iostream>

using std::wstring;
using std::endl;

using namespace CarobNS;

void TestDriverResultSet::testNullValueRetrieval()
{
  wstring fctName(L"TestDriverResultSet::testNullValueRetrieval");
  Statement* statementPtr = NULL;
  //1. Update table: set a value to null
  statementPtr = connectionPtr->createStatement();
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Creating a NULL value");
  }
  statementPtr->executeUpdate(L"UPDATE address SET firstname=NULL WHERE id=0");

  //2. try to read the value with getString
  DriverResultSet* drsPtr = statementPtr->executeQuery(L"SELECT * FROM address WHERE id=0");
  drsPtr->next();
  try
  {
    drsPtr->getString(2); //should fail
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (NullValueException nve)
  {
    if (isErrorEnabled())
    {
      logError(fctName, L"Read failed (this is ok). Exception: "
          + nve.description());
    }
  }
  //3. lets do it again with getAsInt
  try
  {
    drsPtr->getAsInt(2); //should fail
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (NullValueException nve)
  {
    if (isErrorEnabled())
    {
      logError(fctName, L"Read failed (this is ok). Exception: "
          + nve.description());
    }
  }
}

void TestDriverResultSet::testGetAsIntOnString()
{
  wstring fctName(L"TestDriverResultSet::testGetAsIntOnString");
  Statement* statementPtr = NULL;
  //1. Set a string to an int an read it
  statementPtr = connectionPtr->createStatement();
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Writing an int in a string");
  }
  statementPtr->executeUpdate(L"UPDATE address SET firstname='1234' WHERE id=0");
  //read
  DriverResultSet* drsPtr = statementPtr->executeQuery(L"SELECT * FROM address WHERE id=0");
  drsPtr->next();
  CPPUNIT_ASSERT(drsPtr->getAsInt(2) == 1234);

  //2. Negative numbers
  statementPtr = connectionPtr->createStatement();
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Writing a negative int in a string");
  }
  statementPtr->executeUpdate(L"UPDATE address SET firstname='-1' WHERE id=1");
  //Read it
  drsPtr = statementPtr->executeQuery(L"SELECT * FROM address WHERE id=1");
  drsPtr->next();
  CPPUNIT_ASSERT(drsPtr->getAsInt(2) == -1);

  //3. Not a number
  statementPtr = connectionPtr->createStatement();
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Writing a non-number in a string");
  }
  statementPtr->executeUpdate(L"UPDATE address SET firstname='hello' WHERE id=1");
  //Read it
  drsPtr = statementPtr->executeQuery(L"SELECT * FROM address WHERE id=1");
  drsPtr->next();
  try
  {
    drsPtr->getAsInt(2);
    CPPUNIT_ASSERT(false);
  }
  catch (DriverException de)
  {
    if (isErrorEnabled())
    {
      logError(fctName, L"Read failed (this is ok). Exception: "
          + de.description());
    }
  }
}

CppUnit::Test* TestDriverResultSet::suite()
{
  CppUnit::TestSuite *suiteOfTests = new CppUnit::TestSuite( "TestDriverResultSet" );
  suiteOfTests->addTest(new CppUnit::TestCaller<TestDriverResultSet>(
                                 "TestDriverResultSet::testNullValueRetrieval", 
                                 &TestDriverResultSet::testNullValueRetrieval));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestDriverResultSet>(
                                 "TestDriverResultSet::testGetAsIntOnString", 
                                 &TestDriverResultSet::testGetAsIntOnString));
  return suiteOfTests;
}

