/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "TestExecReadRequest.hpp"

#include "Statement.hpp"
#include "ResultSetMetaData.hpp"

#include "CarobException.hpp"
#include "Common.hpp"

#include <string>
#include <iostream>

using std::wstring;
using std::endl;

using namespace CarobNS;

void TestExecReadRequest::testReadBadRequest()
{
  wstring fctName(L"TestExecReadRequest::testReadBadRequest");
  try
  {
    Statement* statementPtr = connectionPtr->createStatement();
    if (isInfoEnabled())
    {
      logInfo(fctName, L"Executing dummy read - should fail");
    }
    statementPtr->executeQuery(L"dummy request");
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (BackendException be)
  {
    if (isErrorEnabled())
    {
      logError(fctName, L"Read failed (this is ok). Exception: "
          + be.description());
    }
  }
}

void TestExecReadRequest::testReadBadTable()
{
  wstring fctName(L"TestExecReadRequest::testReadBadTable");
  try
  {
    Statement* statementPtr = connectionPtr->createStatement();
    if (isInfoEnabled())
    {
      logInfo(fctName, L"Executing bad table read - should fail");
    }
    statementPtr->executeQuery(L"select * from dummy");
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (BackendException be)
  {
    if (isErrorEnabled())
    {
      logError(fctName, L"Read failed (this is ok). Exception: "
          + be.description());
    }
  }
}

void TestExecReadRequest::testReadGood()
{
  wstring fctName(L"TestExecReadRequest::testReadGood");


  Statement* statementPtr = connectionPtr->createStatement();
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Executing read - should succeed");
  }
  statementPtr->setFetchSize(1);
  DriverResultSet* drsPtr = statementPtr->executeQuery(
      L"select * from address");
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Read succeeded. Displaying 50 rows:");
  }
  //Display 50 rows using metadata and toString
  ResultSetMetaData rsmd(drsPtr);
  std::wostringstream buffer;
  buffer<<L"25 first rows got as strings"<<endl;
  for (int i=0; i<25; i++)
  {
    drsPtr->next();
    buffer<<i+1<<L"\t";
    for (int j=0; j<rsmd.getColumnCount(); j++)
      buffer<<drsPtr->getAsString(j+1)<<L"\t";
    buffer<<endl;
  }
  buffer<<L"25 next rows got as Int32, String and AsString"<<endl;
  for (int i=25; i<50; i++)
  {
    drsPtr->next();
    buffer<<i+1<<L"\t"<<drsPtr->getInt32(1)
              <<L"\t"<<drsPtr->getString(2)
              <<L"\t\t"<<drsPtr->getAsString(3)<<endl;
  }
  if (isInfoEnabled())
    logInfo(fctName, buffer.str());
}

void TestExecReadRequest::testReadWithMaxRows()
{
  wstring fctName(L"TestStatement::testReadWithMaxRows");
  Statement* statementPtr = NULL;
  statementPtr = connectionPtr->createStatement();
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Executing read with max rows = 1");
  }
  statementPtr->setMaxRows(1);
  DriverResultSet* drsPtr = statementPtr->executeQuery(
      L"select * from address");
  if (isInfoEnabled())
  {
    logInfo(fctName, L"executeQuery succeeded. Displaying first row.");
  }
  //wcerr<<L"Row\tId\tName\t\tCost"<<endl;
  std::wostringstream buffer;
  buffer<<L"Id\tFirstName\tLastName"<<endl;
  CPPUNIT_ASSERT(drsPtr->next() == true);
  buffer<<drsPtr->getInt32(1)
       <<L"\t"<<drsPtr->getString(2)
       <<L"\t\t"<<drsPtr->getString(3)<<endl;
  if (isInfoEnabled())
  {
    logInfo(fctName, buffer.str());
  }
  //This next() should return false because we asked just 1 row
  CPPUNIT_ASSERT(drsPtr->next() == false);
}

CppUnit::Test* TestExecReadRequest::suite()
{
  CppUnit::TestSuite *suiteOfTests = new CppUnit::TestSuite( "TestExecReadRequest" );
  suiteOfTests->addTest(new CppUnit::TestCaller<TestExecReadRequest>(
                                 "TestExecReadRequest::testReadBadRequest", 
                                 &TestExecReadRequest::testReadBadRequest));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestExecReadRequest>(
                                 "TestExecReadRequest::testReadBadTable", 
                                 &TestExecReadRequest::testReadBadTable));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestExecReadRequest>(
                                 "TestExecReadRequest::testReadGood", 
                                 &TestExecReadRequest::testReadGood));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestExecReadRequest>(
                                 "TestExecReadRequest::testReadWithMaxRows", 
                                 &TestExecReadRequest::testReadWithMaxRows));

  return suiteOfTests;
}
