/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon
 * Contributor(s): Marc Herbert
 */

#include "TestParameterStatement.hpp"

#include "ResultSetMetaData.hpp"
#include "RequestWithResultSetParameters.hpp"
#include "BigDecimal.hpp"
#include "ParameterStatement.hpp"
#include "DriverResultSet.hpp"
#include "TypeConstants.hpp"

#include "Connection.hpp"
#include "CarobException.hpp"
#include "Common.hpp"

#include <iostream>
#include <limits>

using std::wstring;
using std::endl;
using std::numeric_limits;

using namespace CarobNS;

void TestParameterStatement::testExecuteQueryLessParams()
{
  wstring fctName(L"testExecuteQueryLessParams");
  ParameterStatement* statementPtr = NULL;
  try
  {
    statementPtr = connectionPtr->createParameterStatement(L"select * from address where id = ?");
    if (isInfoEnabled())
    {
      logInfo(fctName, L"Executing query without setting it's parameters - should fail");
    }
    statementPtr->executeQuery();
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (DriverException &be)
  {
    if (isErrorEnabled())
    {
      logError(fctName, L"Query failed (this is ok). Exception: "
          + be.description());
    }
  }
}

void TestParameterStatement::testExecuteQueryGood()
{
  wstring fctName(L"TestParameterStatement::testExecuteQueryGood");
  ParameterStatement* statementPtr = NULL;
  statementPtr = connectionPtr->createParameterStatement(L"select * from address where id < ?");
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Executing read - should succeed");
  }
  statementPtr->setInt(1,10);
  statementPtr->setFetchSize(1);
  DriverResultSet* drsPtr = statementPtr->executeQuery();
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Read succeeded. Displaying rows which has id < 10");
  }
  ResultSetMetaData rsmd(drsPtr);
  std::wostringstream buffer;
  int i = 1;
  while (drsPtr->next())
  {
    buffer<<i++<<L"\t";
    for (int j=0; j<rsmd.getColumnCount(); j++)
      buffer<<drsPtr->getAsString(j+1)<<L"\t";
    buffer<<endl;
  }
  if (isInfoEnabled())
    logInfo(fctName, buffer.str());

}

//EXECUTE UPDATE

void TestParameterStatement::testExecuteUpdateBadParams()
{
  wstring fctName(L"testExecuteUpdateBadTable");
  ParameterStatement* statementPtr = NULL;
  try
  {
    statementPtr = connectionPtr->createParameterStatement(L"update address set name=? where id=?");
    if (isInfoEnabled())
    {
      logInfo(fctName, L"Executing update with bad params - this should fail");
    }
    statementPtr->setInt(10,10);
    statementPtr->executeUpdate();
    // We should receive an exception instead of coming here
    CPPUNIT_ASSERT(false);
  }
  catch (DriverException &be)
  {
    if (isErrorEnabled())
    {
      logError(fctName, L"Update failed (this is ok). Exception: "
          + be.description());
    }
  }
}

void TestParameterStatement::testExecuteUpdateGood()
{
  wstring fctName(L"TestParameterStatement::testExecuteUpdateGood");
  ParameterStatement* statementPtr = NULL;
  statementPtr = connectionPtr->createParameterStatement(L"update product set name=? where id=?");
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Executing update - should succeed");
  }
  statementPtr->setString(1, std::wstring(L"changed by testExecuteUpdateGood"));
  statementPtr->setInt(2,0);
  int nbRowsAffected = statementPtr->executeUpdate();
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Update succeeded. Number of affected rows = "
        + toUserString(nbRowsAffected));
  }
  CPPUNIT_ASSERT(nbRowsAffected == 1);
}

void TestParameterStatement::testExecuteWithSelect()
{
  wstring fctName(L"TestParameterStatement::testExecuteWithSelect");
  ParameterStatement* statementPtr = NULL;
  statementPtr = connectionPtr->createParameterStatement(L"select * from address where id < ?");
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Executing execute(SELECT * FROM address where id < 10) - should succeed");
  }
  statementPtr->setInt(1,10);
  bool isRS = statementPtr->execute();
  CPPUNIT_ASSERT(isRS);
  
  DriverResultSet* drsPtr = statementPtr->getResultSet();
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Read succeeded. Displaying rows:");
  }
  ResultSetMetaData rsmd(drsPtr);
  std::wostringstream buffer;
  int i=1;
  while (drsPtr->next())
  {
    buffer<<i++<<L"\t";
    for (int j=0; j<rsmd.getColumnCount(); j++)
      buffer<<drsPtr->getAsString(j+1)<<L"\t";
    buffer<<endl;
  }
  if (isInfoEnabled())
  {
    logInfo(fctName, buffer.str());
  }
}

void TestParameterStatement::testExecuteWithUpdate()
{
  wstring fctName(L"TestParameterStatement::testExecuteWithUpdate");
  ParameterStatement* statementPtr = NULL;
  statementPtr = connectionPtr->createParameterStatement(L"UPDATE product SET name=? WHERE id=?");
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Executing execute(UPDATE FROM address...) - should succeed");
  }
   statementPtr->setString<const wchar_t *>(1, L"changed by testExecuteWithUpdate");
  statementPtr->setInt(2, 0);
  bool isRS = statementPtr->execute();
  CPPUNIT_ASSERT(!isRS);
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Update with execute succeeded. Number of affected rows = "
        + toUserString(statementPtr->getUpdateCount()));
  }
  CPPUNIT_ASSERT(statementPtr->getUpdateCount() == 1);
  //execute again the statement with different parameters
  statementPtr->setInt(2, 1);
  isRS = statementPtr->execute();
  CPPUNIT_ASSERT(!isRS);
  if (isInfoEnabled())
  {
    logInfo(fctName, L"Update with execute succeeded. Number of affected rows = "
        + toUserString(statementPtr->getUpdateCount()));
  }
  CPPUNIT_ASSERT(statementPtr->getUpdateCount() == 1);
}


void TestParameterStatement::testBoolean()
{
  const std::wstring TABLE_TYPE(L"(b1 boolean, b2 boolean)");
  const std::wstring TABLE_NAME(L"booltable");

  wstring fctName(L"TestParameterStatement::testBoolean");
  
  createOrReplaceTable(connectionPtr, TABLE_NAME, TABLE_TYPE);

  ParameterStatement* pstmt = connectionPtr->
    createParameterStatement(std::wstring() + L"INSERT into " + TABLE_NAME + L" values (?, ?)");

  pstmt->setBoolean(1, false);
  pstmt->setBoolean(2, true);
  int uc = pstmt->executeUpdate();
  CPPUNIT_ASSERT(1 == uc);

  pstmt->setBoolean(1, std::wstring(L"das ist falsch"));
  pstmt->setBoolean(2, std::wstring(L"tRue"));
  uc = pstmt->executeUpdate();
  CPPUNIT_ASSERT(1 == uc);

  Statement* stmt = connectionPtr-> createStatement();

  DriverResultSet * drs = stmt->executeQuery(std::wstring() + L"SELECT * from " + TABLE_NAME);
  CPPUNIT_ASSERT(drs->next());
  CPPUNIT_ASSERT(0 == drs->getAsInt(1)); // FIXME: replace by getBoolean()
  CPPUNIT_ASSERT(1 == drs->getAsInt(2));
  CPPUNIT_ASSERT(drs->next());
  CPPUNIT_ASSERT(0 == drs->getAsInt(1)); // FIXME: replace by getBoolean()
  CPPUNIT_ASSERT(1 == drs->getAsInt(2));


}

void TestParameterStatement::testStrings()
{
  const std::wstring TABLE_TYPE(L"(str varchar(100))");
  const std::wstring TABLE_NAME(L"stringtable");

  wstring fctName(L"TestParameterStatement::testStrings");
  
  createOrReplaceTable(connectionPtr, TABLE_NAME, TABLE_TYPE);

  ParameterStatement* pstmt = connectionPtr->
    createParameterStatement(wstring() + L"INSERT into " + TABLE_NAME + L" values (?)");

  wstring teststring(L"|!%>");

  pstmt->setString(1, teststring);
  int uc = pstmt->executeUpdate();
  CPPUNIT_ASSERT(1 == uc);

  Statement* stmt = connectionPtr-> createStatement();

  DriverResultSet * drs = stmt->executeQuery(std::wstring() + L"SELECT * from " + TABLE_NAME);
  CPPUNIT_ASSERT(drs->next());
  CPPUNIT_ASSERT(teststring == drs->getString(1)); // FIXME: replace by getBoolean()
}

void TestParameterStatement::testBigDecimal()
{
  const std::wstring TABLE_TYPE(L"(bd DECIMAL(128))");
  const std::wstring TABLE_NAME(L"bigdecimaltest");

  wstring fctName(L"TestParameterStatement::testBigDecimal");
  
  createOrReplaceTable(connectionPtr, TABLE_NAME, TABLE_TYPE);

  ParameterStatement* pstmt = connectionPtr->
    createParameterStatement(wstring(L"INSERT into ") + TABLE_NAME + L" values (?)");

  BigDecimal testVals[15];
  int nbTestVals = 0;
  testVals[nbTestVals++] = BigDecimal(L"-123,456");
  testVals[nbTestVals++] = BigDecimal(L"123,456");
  testVals[nbTestVals++] = BigDecimal(-123456789);
  testVals[nbTestVals++] = BigDecimal(123456789);
  testVals[nbTestVals++] = BigDecimal(static_cast<int64_t>(-123456789L));
  testVals[nbTestVals++] = BigDecimal(static_cast<int64_t>(123456789L));
  testVals[nbTestVals++] = BigDecimal(static_cast<int64_t>(-123456789LL));
  testVals[nbTestVals++] = BigDecimal(static_cast<int64_t>(123456789LL));
  testVals[nbTestVals++] = BigDecimal(numeric_limits<int32_t>::min());
  testVals[nbTestVals++] = BigDecimal(numeric_limits<int32_t>::max());
  testVals[nbTestVals++] = BigDecimal(numeric_limits<int64_t>::min());
  testVals[nbTestVals++] = BigDecimal(numeric_limits<int64_t>::max());
  testVals[nbTestVals++] = BigDecimal(static_cast<int64_t>(numeric_limits<long long>::min()));
  testVals[nbTestVals++] = BigDecimal(static_cast<int64_t>(numeric_limits<long long>::max()));
  //TODO change testVals array size when adding new values

  for (int i=0; i<nbTestVals; i++)
  {
    logDebug(fctName, L"setting value " + static_cast<wstring>(testVals[i]));
    pstmt->setBigDecimal(1, testVals[i]);
    int uc = pstmt->executeUpdate();
    CPPUNIT_ASSERT(uc == 1);
  }
  Statement* stmt = connectionPtr-> createStatement();

  DriverResultSet * drs = stmt->executeQuery(std::wstring(L"SELECT * from ") + TABLE_NAME);
  for (int i=0; i<nbTestVals; i++)
  {
    CPPUNIT_ASSERT(drs->next());
    logDebug(fctName, L"testing with value " + static_cast<wstring>(testVals[i]));
    CPPUNIT_ASSERT(drs->getBigDecimal(1) == testVals[i]);
  }
}

void TestParameterStatement::testSetNull()
{
  const std::wstring TABLE_TYPE(L"(c1 smallint, c2 smallint)");
  const std::wstring TABLE_NAME(L"testSetNull");

  wstring fctName(L"TestParameterStatement::testSetNull");
  
  createOrReplaceTable(connectionPtr, TABLE_NAME, TABLE_TYPE);

  ParameterStatement* pstmt = connectionPtr->
    createParameterStatement(wstring(L"INSERT into ") + TABLE_NAME + L" values (?, ?)");
  pstmt->setInt(1, -125);
  pstmt->setNull(2, SQLT_SMALLINT);
  pstmt->execute();
  Statement* stmt = connectionPtr-> createStatement();
  DriverResultSet* drs = stmt->executeQuery(std::wstring(L"SELECT * from ") + TABLE_NAME);
  CPPUNIT_ASSERT(drs->next());
  try
  {
    logDebug(fctName, L"Trying to get null value - should throw an exception");
    drs->getInt32(2);
    CPPUNIT_ASSERT(false);
  }
  catch (NullValueException nve)
  {
    logDebug(fctName, L"NullValueException caught - this is ok");
  }
}

CppUnit::Test* TestParameterStatement::suite()
{
  CppUnit::TestSuite *suiteOfTests = new CppUnit::TestSuite( "TestParameterStatement" );
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testExecuteQueryLessParams", 
                                 &TestParameterStatement::testExecuteQueryLessParams));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testExecuteQueryGood", 
                                 &TestParameterStatement::testExecuteQueryGood));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testExecuteUpdateBadParams", 
                                 &TestParameterStatement::testExecuteUpdateBadParams));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testExecuteUpdateGood", 
                                 &TestParameterStatement::testExecuteUpdateGood));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testExecuteWithSelect", 
                                 &TestParameterStatement::testExecuteWithSelect));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testExecuteWithUpdate", 
                                 &TestParameterStatement::testExecuteWithUpdate));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testBoolean", 
                                 &TestParameterStatement::testBoolean));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testStrings",
                                 &TestParameterStatement::testStrings));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testBigDecimal",
                                 &TestParameterStatement::testBigDecimal));
  suiteOfTests->addTest(new CppUnit::TestCaller<TestParameterStatement>(
                                 "TestParameterStatement::testSetNull",
                                 &TestParameterStatement::testSetNull));

  return suiteOfTests;
}
