/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "Common.hpp"
#include "ConnectionParameters.hpp"
#include "CarobException.hpp"
#include "Statement.hpp"

#include "ConnectionSetup.hpp"

#include <vector>

using std::wstring;

using namespace CarobNS;

const std::wstring ConnectionSetup::DEFAULT_HOST = L"localhost";
const int ConnectionSetup::DEFAULT_PORT = 25322;

void ConnectionSetup::setUp()
{
  std::vector<ControllerInfo> ctrls;
#if 1
  ctrls.push_back(ControllerInfo(DEFAULT_HOST, DEFAULT_PORT));
  ctrls.push_back(ControllerInfo(DEFAULT_HOST, 25323));
  ConnectionParameters connectionPrms(ctrls,
                                      ConnectionParameters::DEFAULT_DB,
                                      ConnectionParameters::DEFAULT_USER,
                                      ConnectionParameters::DEFAULT_PASSWD,
                                      ConnectionParameters::DEFAULT_POLICY,
                                      ConnectionParameters::DEFAULT_CONNECTION_PERSISTENCY,
                                      ConnectionParameters::DEFAULT_RETRIEVE_SQL_WARNINGS);
#else
  // Put your controller settings here.
  // Check this constructor for more details.
  // TODO: these should be run-time settings.
  ctrls.push_back(Controller(L"myhostname", 25322));
  ConnectionParameters connectionPrms(ctrls,
                                      ConnectionParameters::DEFAULT_DB,
                                      ConnectionParameters::DEFAULT_USER,
                                      ConnectionParameters::DEFAULT_PASSWD,
                                      ConnectionParameters::DEFAULT_POLICY,
                                      ConnectionParameters::DEFAULT_CONNECTION_PERSISTENCY,
                                      ConnectionParameters::DEFAULT_RETRIEVE_SQL_WARNINGS);
#endif
  try
  {
    connectionPtr = new Connection(connectionPrms);
  }
  catch (CarobException e)
  {
    logError(L"ConnectionSetup::setUp()", L"Connection failed: " + e.description()); 
    connectionPtr = NULL;
    CPPUNIT_ASSERT_MESSAGE("Could not connect to default host", false);
  }
}

void ConnectionSetup::tearDown()
{
  if (connectionPtr)
    delete connectionPtr;
}

void dropTableIfExists(Connection *conn_p, const std::wstring& tableName)
{
    std::wstring dropcommand(L"drop table ");
    dropcommand += tableName;

    Statement* dropstmt = conn_p->createStatement();

    try {

        int updatecount = dropstmt->executeUpdate(dropcommand);
        CPPUNIT_ASSERT(0 == updatecount);

    } catch (const BackendException& be)
    {
        // rethrow if different from UNDEFINED TABLE
        // TODO: ignore case here. Problem: "case" concept is locale sensitive?
        if (0 != be.getSQLState().compare(L"42P01") // postgres
            && 0 != be.getSQLState().compare(L"S0002")) // hypersonic
            throw be; // and leak. Who cares?
        // else ignore
    }

    conn_p->deleteStatement(dropstmt);
}

void createOrReplaceTable(Connection *conn_p, const std::wstring& tableName, const std::wstring& tableType)
{
    dropTableIfExists(conn_p, tableName);

    std::wstring createcommand(L"create table ");
    createcommand += tableName + L" ";
    createcommand += tableType;
    
    Statement* createstmt = conn_p->createStatement();

    int updatecount = createstmt->executeUpdate(createcommand);
    CPPUNIT_ASSERT(0 == updatecount);

    conn_p->deleteStatement(createstmt);
}


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
